import numpy as np
from uuid import uuid4
from datetime import datetime

import pynwb
from hdmf.testing import TestCase

from nwbinspector.tools import all_of_type
from nwbinspector.inspector_tools import organize_messages
from nwbinspector.register_checks import InspectorMessage, Importance, Severity


def test_all_of_type():
    nwbfile = pynwb.NWBFile(
        session_description="Testing inspector.",
        identifier=str(uuid4()),
        session_start_time=datetime.now().astimezone(),
    )
    true_time_series = [
        pynwb.TimeSeries(name=f"time_series_{x}", data=np.zeros(shape=(100, 10)), rate=1.0, unit="") for x in range(4)
    ]
    for x in range(2):
        nwbfile.add_acquisition(true_time_series[x])
    ecephys_module = nwbfile.create_processing_module(name="ecephys", description="")
    ecephys_module.add(true_time_series[2])
    ophys_module = nwbfile.create_processing_module(name="ophys", description="")
    ophys_module.add(true_time_series[3])

    nwbfile_time_series = [obj for obj in all_of_type(nwbfile=nwbfile, neurodata_type=pynwb.TimeSeries)]
    for time_series in true_time_series:
        assert time_series in nwbfile_time_series


class TestOrganization(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.messages = [
            InspectorMessage(
                message="test1",
                importance=Importance.BEST_PRACTICE_SUGGESTION,
                check_function_name="fun1",
                object_type="ElectricalSeries",
                object_name="ts1",
                location="/acquisition/",
                file_path="path1/file1.nwb",
            ),
            InspectorMessage(
                message="test2",
                importance=Importance.CRITICAL,
                check_function_name="fun2",
                object_type="DynamicTable",
                object_name="tab",
                location="/acquisition/",
                file_path="path1/file1.nwb",
            ),
            InspectorMessage(
                message="test3",
                importance=Importance.BEST_PRACTICE_SUGGESTION,
                severity=Severity.HIGH,
                check_function_name="fun3",
                object_type="NWBFile",
                object_name="root",
                location="/",
                file_path="path2/file2.nwb",
            ),
            InspectorMessage(
                message="test4",
                importance=Importance.CRITICAL,
                check_function_name="fun2",
                object_type="ElectricalSeries",
                object_name="ts1",
                location="/processing/ecephys/LFP/",
                file_path="path2/path3/file3.nwb",
            ),
        ]

    def test_message_level_assertion(self):
        for level in ["message", "object_name", "severity"]:
            with self.assertRaisesWith(
                exc_type=AssertionError,
                exc_msg=(
                    "You must specify levels to organize by that correspond to attributes of the InspectorMessage "
                    "class, excluding the text message, object_name, and severity."
                ),
            ):
                organize_messages(messages=self.messages, levels=[level])

    def test_file_by_importance(self):
        test_result = organize_messages(messages=self.messages, levels=["file_path", "importance"])
        true_result = {
            "path1/file1.nwb": {
                Importance.CRITICAL: [
                    InspectorMessage(
                        message="test2",
                        importance=Importance.CRITICAL,
                        severity=Severity.LOW,
                        check_function_name="fun2",
                        object_type="DynamicTable",
                        object_name="tab",
                        location="/acquisition/",
                        file_path="path1/file1.nwb",
                    )
                ],
                Importance.BEST_PRACTICE_SUGGESTION: [
                    InspectorMessage(
                        message="test1",
                        importance=Importance.BEST_PRACTICE_SUGGESTION,
                        severity=Severity.LOW,
                        check_function_name="fun1",
                        object_type="ElectricalSeries",
                        object_name="ts1",
                        location="/acquisition/",
                        file_path="path1/file1.nwb",
                    )
                ],
            },
            "path2/file2.nwb": {
                Importance.BEST_PRACTICE_SUGGESTION: [
                    InspectorMessage(
                        message="test3",
                        importance=Importance.BEST_PRACTICE_SUGGESTION,
                        severity=Severity.HIGH,
                        check_function_name="fun3",
                        object_type="NWBFile",
                        object_name="root",
                        location="/",
                        file_path="path2/file2.nwb",
                    )
                ]
            },
            "path2/path3/file3.nwb": {
                Importance.CRITICAL: [
                    InspectorMessage(
                        message="test4",
                        importance=Importance.CRITICAL,
                        severity=Severity.LOW,
                        check_function_name="fun2",
                        object_type="ElectricalSeries",
                        object_name="ts1",
                        location="/processing/ecephys/LFP/",
                        file_path="path2/path3/file3.nwb",
                    )
                ]
            },
        }
        self.assertDictEqual(d1=test_result, d2=true_result)

    def test_reverse(self):
        test_result = organize_messages(
            messages=self.messages, levels=["importance", "file_path"], reverse=[False, True]
        )
        true_result = {
            Importance.CRITICAL: {
                "path2/path3/file3.nwb": [
                    InspectorMessage(
                        message="test4",
                        importance=Importance.CRITICAL,
                        severity=Severity.LOW,
                        check_function_name="fun2",
                        object_type="ElectricalSeries",
                        object_name="ts1",
                        location="/processing/ecephys/LFP/",
                        file_path="path2/path3/file3.nwb",
                    )
                ],
                "path1/file1.nwb": [
                    InspectorMessage(
                        message="test2",
                        importance=Importance.CRITICAL,
                        severity=Severity.LOW,
                        check_function_name="fun2",
                        object_type="DynamicTable",
                        object_name="tab",
                        location="/acquisition/",
                        file_path="path1/file1.nwb",
                    )
                ],
            },
            Importance.BEST_PRACTICE_SUGGESTION: {
                "path2/file2.nwb": [
                    InspectorMessage(
                        message="test3",
                        importance=Importance.BEST_PRACTICE_SUGGESTION,
                        severity=Severity.HIGH,
                        check_function_name="fun3",
                        object_type="NWBFile",
                        object_name="root",
                        location="/",
                        file_path="path2/file2.nwb",
                    )
                ],
                "path1/file1.nwb": [
                    InspectorMessage(
                        message="test1",
                        importance=Importance.BEST_PRACTICE_SUGGESTION,
                        severity=Severity.LOW,
                        check_function_name="fun1",
                        object_type="ElectricalSeries",
                        object_name="ts1",
                        location="/acquisition/",
                        file_path="path1/file1.nwb",
                    )
                ],
            },
        }
        self.assertDictEqual(d1=test_result, d2=true_result)



import pytest
from packaging.version import Version
from pynwb.icephys import IntracellularElectrode
from pynwb.device import Device

from nwbinspector import InspectorMessage, Importance, check_intracellular_electrode_cell_id_exists
from nwbinspector.utils import get_package_version

PYNWB_VERSION_LOWER_2_1_0 = get_package_version(name="pynwb") < Version("2.1.0")
PYNWB_VERSION_LOWER_SKIP_REASON = "This test requires PyNWB>=2.1.0"


@pytest.mark.skipif(PYNWB_VERSION_LOWER_2_1_0, reason=PYNWB_VERSION_LOWER_SKIP_REASON)
def test_pass_check_intracellular_electrode_cell_id_exists():
    device = Device(name="device")
    ielec = IntracellularElectrode(name="ielec", cell_id="123", device=device, description="an intracellular electrode")
    assert check_intracellular_electrode_cell_id_exists(ielec) is None


@pytest.mark.skipif(PYNWB_VERSION_LOWER_2_1_0, reason=PYNWB_VERSION_LOWER_SKIP_REASON)
def test_fail_check_intracellular_electrode_cell_id_exists():
    device = Device(name="device")
    ielec = IntracellularElectrode(name="ielec", device=device, description="an intracellular electrode")
    assert check_intracellular_electrode_cell_id_exists(ielec) == InspectorMessage(
        message="Please include a unique cell_id associated with this IntracellularElectrode.",
        importance=Importance.BEST_PRACTICE_VIOLATION,
        check_function_name="check_intracellular_electrode_cell_id_exists",
        object_type="IntracellularElectrode",
        object_name="ielec",
        location="/",
    )


@pytest.mark.skipif(not PYNWB_VERSION_LOWER_2_1_0, reason="This test requires PyNWB<2.1.0")
def test_skip_check_for_lower_versions():
    device = Device(name="device")
    ielec = IntracellularElectrode(name="ielec", device=device, description="an intracellular electrode")
    assert check_intracellular_electrode_cell_id_exists(ielec) is None



"""Checks specific to the Images neurodata type."""
from ..register_checks import register_check, Importance, InspectorMessage

try:
    from pynwb.base import Images

    HAVE_IMAGES = True
except ImportError:
    HAVE_IMAGES = False


if HAVE_IMAGES:

    @register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=Images)
    def check_order_of_images_unique(images: Images):
        """Check that all the values in the order_of_images field of an Images object are unique."""
        if images.order_of_images is None:
            return
        if not len(set(images.order_of_images)) == len(images.order_of_images):
            return InspectorMessage(message="order_of_images should have unique values.")

    @register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=Images)
    def check_order_of_images_len(images: Images):
        """Check that all the values in the order_of_images field of an Images object are unique."""
        if images.order_of_images is None:
            return
        if not len(images.order_of_images) == len(images.images):
            return InspectorMessage(
                message=f"Length of order_of_images ({len(images.order_of_images)}) does not match the number of "
                f"images ({len(images.images)})."
            )



"""Check functions that can apply to any descendant of DynamicTable."""
from numbers import Real
from typing import List, Optional

import numpy as np
from hdmf.common import DynamicTable, DynamicTableRegion, VectorIndex
from hdmf.utils import get_data_shape
from pynwb.file import TimeIntervals, Units

from ..register_checks import register_check, InspectorMessage, Importance
from ..utils import (
    _cache_data_selection,
    format_byte_size,
    is_ascending_series,
    is_dict_in_string,
    is_string_json_loadable,
)


@register_check(importance=Importance.CRITICAL, neurodata_type=DynamicTableRegion)
def check_dynamic_table_region_data_validity(dynamic_table_region: DynamicTableRegion, nelems=200):
    """Check if a DynamicTableRegion is valid."""
    if np.any(np.asarray(dynamic_table_region.data[:nelems]) > len(dynamic_table_region.table)):
        return InspectorMessage(
            message=(
                f"Some elements of {dynamic_table_region.name} are out of range because they are greater than the "
                "length of the target table. Note that data should contain indices, not ids."
            )
        )
    if np.any(np.asarray(dynamic_table_region.data[:nelems]) < 0):
        return InspectorMessage(
            message=f"Some elements of {dynamic_table_region.name} are out of range because they are less than 0."
        )


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=DynamicTable)
def check_empty_table(table: DynamicTable):
    """Check if a DynamicTable is empty."""
    if len(table.id) == 0:
        return InspectorMessage(message="This table has no data added to it.")


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=TimeIntervals)
def check_time_interval_time_columns(time_intervals: TimeIntervals, nelems: int = 200):
    """
    Check that time columns are in ascending order.

    Parameters
    ----------
    time_intervals: TimeIntervals
    nelems: int
        Only check the first {nelems} elements. This is useful in case there columns are
        very long so you don't need to load the entire array into memory. Use None to
        load the entire arrays.
    """
    unsorted_cols = []
    for column in time_intervals.columns:
        if column.name[-5:] == "_time":
            if not is_ascending_series(column.data, nelems):
                unsorted_cols.append(column.name)
    if unsorted_cols:
        return InspectorMessage(
            message=(
                f"{unsorted_cols} are time columns but the values are not in ascending order. "
                "All times should be in seconds with respect to the session start time."
            )
        )


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=TimeIntervals)
def check_time_intervals_stop_after_start(time_intervals: TimeIntervals, nelems: int = 200):
    """
    Check that all stop times on a TimeInterval object occur after their corresponding start times.

    Best Practice: :ref:`best_practice_time_interval_time_columns`

    Parameters
    ----------
    time_intervals: TimeIntervals
    nelems: int
        Only check the first {nelems} elements. This is useful in case there columns are
        very long so you don't need to load the entire array into memory. Use None to
        load the entire arrays.
    """
    if np.any(
        np.asarray(_cache_data_selection(data=time_intervals["stop_time"].data, selection=slice(nelems)))
        - np.asarray(_cache_data_selection(data=time_intervals["start_time"].data, selection=slice(nelems)))
        < 0
    ):
        return InspectorMessage(
            message=(
                "stop_times should be greater than start_times. Make sure the stop times are with respect to the "
                "session start time."
            )
        )


@register_check(importance=Importance.BEST_PRACTICE_SUGGESTION, neurodata_type=DynamicTable)
def check_column_binary_capability(table: DynamicTable, nelems: int = 200):
    """
    Check each column of a table to see if the data could be set as a boolean dtype.

    Parameters
    ----------
    time_intervals: DynamicTable
    nelems: int
        Only check the first {nelems} elements. This is useful in case there columns are
        very long so you don't need to load the entire array into memory. Use None to
        load the entire arrays.
    """
    for column in table.columns:
        if hasattr(column, "data") and not isinstance(column, VectorIndex):
            if np.asarray(column.data[0]).itemsize == 1:
                continue  # already boolean, int8, or uint8
            try:
                unique_values = np.unique(_cache_data_selection(data=column.data, selection=slice(nelems)))
            except TypeError:  # some contained objects are unhashable or have no comparison defined
                continue
            if unique_values.size != 2:
                continue
            parsed_unique_values = np.array(unique_values)
            if isinstance(parsed_unique_values[0], Real):  # upcast to float for comparison
                parsed_unique_values = parsed_unique_values.astype(float)
            elif str(parsed_unique_values.dtype)[:2] == "<U":  # parse strings as all lower-case
                for j in range(2):
                    parsed_unique_values[j] = parsed_unique_values[j].lower()
            pairs_to_check = [
                [1.0, 0.0],
                ["yes", "no"],
                ["y", "n"],
                ["true", "false"],
                ["t", "f"],
                ["hit", "miss"],
            ]
            if any([set(parsed_unique_values) == set(pair) for pair in pairs_to_check]):
                saved_bytes = (unique_values.dtype.itemsize - 1) * np.product(
                    get_data_shape(data=column.data, strict_no_data_load=True)
                )
                if unique_values.dtype == "float":
                    print_dtype = "floats"
                elif unique_values.dtype == "int":
                    print_dtype = "integers"
                elif str(unique_values.dtype)[:2] == "<U":
                    print_dtype = "strings"
                else:
                    print_dtype = f"{unique_values.dtype}"
                yield InspectorMessage(
                    message=(
                        f"Column '{column.name}' uses '{print_dtype}' but has binary values {unique_values}. Consider "
                        "making it boolean instead and renaming the column to start with 'is_'; doing so will "
                        f"save {format_byte_size(byte_size=saved_bytes)}."
                    )
                )


@register_check(importance=Importance.BEST_PRACTICE_SUGGESTION, neurodata_type=DynamicTable)
def check_single_row(
    table: DynamicTable,
    exclude_types: Optional[list] = (Units,),
    exclude_names: Optional[List[str]] = ("electrodes",),
):
    """
    Check if DynamicTable has only a single row; may be better represented by another data type.

    Skips the Units table since it is OK to have only a single spiking unit.
    Skips the Electrode table since it is OK to have only a single electrode.
    """
    if any((isinstance(table, exclude_type) for exclude_type in exclude_types)):
        return
    if any((table.name == exclude_name for exclude_name in exclude_names)):
        return
    if len(table.id) == 1:
        return InspectorMessage(
            message="This table has only a single row; it may be better represented by another data type."
        )


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=DynamicTable)
def check_table_values_for_dict(table: DynamicTable, nelems: int = 200):
    """Check if any values in a row or column of a table contain a string casting of a Python dictionary."""
    for column in table.columns:
        if not hasattr(column, "data") or isinstance(column, VectorIndex) or not isinstance(column.data[0], str):
            continue
        for string in _cache_data_selection(data=column.data, selection=slice(nelems)):
            if is_dict_in_string(string=string):
                message = (
                    f"The column '{column.name}' contains a string value that contains a dictionary! Please "
                    "unpack dictionaries as additional rows or columns of the table."
                )
                if is_string_json_loadable(string=string):
                    message += " This string is also JSON loadable, so call `json.loads(...)` on the string to unpack."
                yield InspectorMessage(message=message)


@register_check(importance=Importance.BEST_PRACTICE_SUGGESTION, neurodata_type=DynamicTable)
def check_col_not_nan(table: DynamicTable, nelems: Optional[int] = 200):
    """Check if all of the values in a single column of a table are NaN."""
    for column in table.columns:
        if (
            not hasattr(column, "data")
            or isinstance(column, VectorIndex)
            or not np.issubdtype(np.array(column[0]).dtype, np.floating)
        ):
            continue
        if nelems is not None and not all(np.isnan(column[:nelems]).flatten()):
            continue

        slice_by = np.ceil(len(column.data) / nelems).astype(int) if nelems else None
        message = f"Column '{column.name}' might have all NaN values. Consider removing it from the table."
        message = message.replace("might have", "has") if nelems is None or slice_by == 1 else message
        if all(np.isnan(column[slice(0, None, slice_by)]).flatten()):
            yield InspectorMessage(message=message)



import numpy as np
import pynwb
import pytest
from packaging import version

from nwbinspector import (
    InspectorMessage,
    Importance,
    check_regular_timestamps,
    check_data_orientation,
    check_timestamps_match_first_dimension,
    check_timestamps_ascending,
    check_missing_unit,
    check_resolution,
)
from nwbinspector.testing import check_streaming_tests_enabled
from nwbinspector.utils import get_package_version, robust_s3_read

STREAMING_TESTS_ENABLED, DISABLED_STREAMING_TESTS_REASON = check_streaming_tests_enabled()


def test_check_regular_timestamps():
    assert check_regular_timestamps(
        time_series=pynwb.TimeSeries(
            name="test_time_series",
            unit="test_units",
            data=np.zeros(shape=3),
            timestamps=[1.2, 3.2, 5.2],
        )
    ) == InspectorMessage(
        message=(
            "TimeSeries appears to have a constant sampling rate. Consider specifying starting_time=1.2 and rate=2.0 "
            "instead of timestamps."
        ),
        importance=Importance.BEST_PRACTICE_VIOLATION,
        check_function_name="check_regular_timestamps",
        object_type="TimeSeries",
        object_name="test_time_series",
        location="/",
    )


def test_pass_check_regular_timestamps():
    """Should pass because there are only two timestamps"""
    assert (
        check_regular_timestamps(
            time_series=pynwb.TimeSeries(
                name="test_time_series",
                unit="test_units",
                data=[0, 0],
                timestamps=[1.2, 3.2],
            )
        )
        is None
    )


def test_check_data_orientation():
    assert check_data_orientation(
        time_series=pynwb.TimeSeries(
            name="test_time_series",
            unit="test_units",
            data=np.zeros(shape=(2, 100)),
            rate=1.0,
        )
    ) == InspectorMessage(
        message=(
            "Data may be in the wrong orientation. "
            "Time should be in the first dimension, and is usually the longest dimension. "
            "Here, another dimension is longer."
        ),
        importance=Importance.CRITICAL,
        check_function_name="check_data_orientation",
        object_type="TimeSeries",
        object_name="test_time_series",
        location="/",
    )


def test_check_timestamps():
    assert check_timestamps_match_first_dimension(
        time_series=pynwb.TimeSeries(
            name="test_time_series",
            unit="test_units",
            data=np.zeros(shape=4),
            timestamps=[1.0, 2.0, 3.0],
        )
    ) == InspectorMessage(
        message="The length of the first dimension of data does not match the length of timestamps.",
        importance=Importance.CRITICAL,
        check_function_name="check_timestamps_match_first_dimension",
        object_type="TimeSeries",
        object_name="test_time_series",
        location="/",
    )


def test_check_timestamps_empty_data():
    assert check_timestamps_match_first_dimension(
        time_series=pynwb.TimeSeries(name="test_time_series", unit="test_units", data=[], timestamps=[1.0, 2.0, 3.0])
    ) == InspectorMessage(
        message="The length of the first dimension of data does not match the length of timestamps.",
        importance=Importance.CRITICAL,
        check_function_name="check_timestamps_match_first_dimension",
        object_type="TimeSeries",
        object_name="test_time_series",
        location="/",
    )


def test_check_timestamps_empty_timestamps():
    assert check_timestamps_match_first_dimension(
        time_series=pynwb.TimeSeries(name="test_time_series", unit="test_units", data=np.zeros(shape=4), timestamps=[])
    ) == InspectorMessage(
        message="The length of the first dimension of data does not match the length of timestamps.",
        importance=Importance.CRITICAL,
        check_function_name="check_timestamps_match_first_dimension",
        object_type="TimeSeries",
        object_name="test_time_series",
        location="/",
    )


def test_pass_check_timestamps_ascending_pass():
    time_series = pynwb.TimeSeries(name="test_time_series", unit="test_units", data=[1, 2, 3], timestamps=[1, 2, 3])
    assert check_timestamps_ascending(time_series) is None


def test_check_timestamps_ascending_fail():
    time_series = pynwb.TimeSeries(name="test_time_series", unit="test_units", data=[1, 2, 3], timestamps=[1, 3, 2])
    assert check_timestamps_ascending(time_series) == InspectorMessage(
        message="test_time_series timestamps are not ascending.",
        importance=Importance.BEST_PRACTICE_VIOLATION,
        check_function_name="check_timestamps_ascending",
        object_type="TimeSeries",
        object_name="test_time_series",
        location="/",
    )


def test_check_missing_unit_pass():
    time_series = pynwb.TimeSeries(name="test_time_series", unit="test_units", data=[1, 2, 3], timestamps=[1, 2, 3])
    assert check_missing_unit(time_series) is None


def test_check_missing_unit_fail():
    time_series = pynwb.TimeSeries(name="test_time_series", unit="", data=[1, 2, 3], timestamps=[1, 2, 3])
    assert check_missing_unit(time_series) == InspectorMessage(
        message="Missing text for attribute 'unit'. Please specify the scientific unit of the 'data'.",
        importance=Importance.BEST_PRACTICE_VIOLATION,
        check_function_name="check_missing_unit",
        object_type="TimeSeries",
        object_name="test_time_series",
        location="/",
    )


def test_check_positive_resolution_pass():
    time_series = pynwb.TimeSeries(name="test", unit="test_units", data=[1, 2, 3], timestamps=[1, 2, 3], resolution=3.4)
    assert check_resolution(time_series) is None


def test_check_unknown_resolution_pass():
    for valid_unknown in [-1.0, np.nan]:
        time_series = pynwb.TimeSeries(name="test", unit="test", data=[1], timestamps=[1], resolution=valid_unknown)
        assert check_resolution(time_series) is None


@pytest.mark.skipif(
    not STREAMING_TESTS_ENABLED or get_package_version("hdmf") >= version.parse("3.3.1"),
    reason=f"{DISABLED_STREAMING_TESTS_REASON or ''}. Also needs 'hdmf<3.3.1'.",
)
def test_check_none_matnwb_resolution_pass():
    """
    Special test on the original problematic file found at

    https://dandiarchive.org/dandiset/000065/draft/files?location=sub-Kibbles%2F

    produced with MatNWB, when read with PyNWB~=2.0.1 and HDMF<=3.2.1 contains a resolution value of None.
    """
    with pynwb.NWBHDF5IO(
        path="https://dandiarchive.s3.amazonaws.com/blobs/da5/107/da510761-653e-4b81-a330-9cdae4838180",
        mode="r",
        load_namespaces=True,
        driver="ros3",
    ) as io:
        nwbfile = robust_s3_read(command=io.read)
        time_series = robust_s3_read(
            command=nwbfile.processing["video_files"]["video"].time_series.get,
            command_args=["20170203_KIB_01_s1.1.h264"],
        )
    assert check_resolution(time_series) is None


def test_check_resolution_fail():
    time_series = pynwb.TimeSeries(name="test", unit="test", data=[1, 2, 3], timestamps=[1, 2, 3], resolution=-2.0)
    assert check_resolution(time_series) == InspectorMessage(
        message="'resolution' should use -1.0 or NaN for unknown instead of -2.0.",
        importance=Importance.BEST_PRACTICE_VIOLATION,
        check_function_name="check_resolution",
        object_type="TimeSeries",
        object_name="test",
        location="/",
    )



"""Primary functions for inspecting NWBFiles."""
import os
import re
import importlib
import traceback
import json
import jsonschema
from pathlib import Path
from enum import Enum
from typing import Union, Optional, List
from concurrent.futures import ProcessPoolExecutor, as_completed
from types import FunctionType
from warnings import filterwarnings, warn
from distutils.util import strtobool
from collections import defaultdict

import click
import pynwb
import yaml
from tqdm import tqdm
from natsort import natsorted

from . import available_checks
from .inspector_tools import (
    get_report_header,
    format_messages,
    print_to_console,
    save_report,
)
from .register_checks import InspectorMessage, Importance
from .tools import get_s3_urls_and_dandi_paths
from .utils import FilePathType, PathType, OptionalListOfStrings, robust_s3_read, calculate_number_of_cpu

INTERNAL_CONFIGS = dict(dandi=Path(__file__).parent / "internal_configs" / "dandi.inspector_config.yaml")


class InspectorOutputJSONEncoder(json.JSONEncoder):
    """Custom JSONEncoder for the NWBInspector."""

    def default(self, o):  # noqa D102
        if isinstance(o, InspectorMessage):
            return o.__dict__
        if isinstance(o, Enum):
            return o.name
        else:
            return super().default(o)


def validate_config(config: dict):
    """Validate an instance of configuration against the official schema."""
    with open(file=Path(__file__).parent / "config.schema.json", mode="r") as fp:
        schema = json.load(fp=fp)
    jsonschema.validate(instance=config, schema=schema)


def _copy_function(function):
    """
    Copy the core parts of a given function, excluding wrappers, then return a new function.

    Based off of
    https://stackoverflow.com/questions/6527633/how-can-i-make-a-deepcopy-of-a-function-in-python/30714299#30714299
    """
    copied_function = FunctionType(
        function.__code__, function.__globals__, function.__name__, function.__defaults__, function.__closure__
    )

    # in case f was given attrs (note this dict is a shallow copy)
    copied_function.__dict__.update(function.__dict__)
    return copied_function


def copy_check(check):
    """
    Copy a check function so that internal attributes can be adjusted without changing the original function.

    Required to ensure our configuration of functions in the registry does not effect the registry itself.

    Also copies the wrapper for auto-parsing results,
    see https://github.com/NeurodataWithoutBorders/nwbinspector/pull/218 for explanation.

    Taken from
    https://stackoverflow.com/questions/6527633/how-can-i-make-a-deepcopy-of-a-function-in-python/30714299#30714299
    """
    copied_check = _copy_function(function=check)
    copied_check.__wrapped__ = _copy_function(function=check.__wrapped__)
    return copied_check


def load_config(filepath_or_keyword: PathType) -> dict:
    """
    Load a config dictionary either via keyword search of the internal configs, or an explicit filepath.

    Currently supported keywords are:
        - 'dandi'
            For all DANDI archive related practices, including validation and upload.
    """
    file = INTERNAL_CONFIGS.get(filepath_or_keyword, filepath_or_keyword)
    with open(file=file, mode="r") as stream:
        config = yaml.load(stream=stream, Loader=yaml.Loader)
    return config


def configure_checks(
    checks: list = available_checks,
    config: Optional[dict] = None,
    ignore: Optional[List[str]] = None,
    select: Optional[List[str]] = None,
    importance_threshold: Importance = Importance.BEST_PRACTICE_SUGGESTION,
) -> list:
    """
    Filter a list of check functions (the entire base registry by default) according to the configuration.

    Parameters
    ----------
    config : dict
        Dictionary valid against our JSON configuration schema.
        Can specify a mapping of importance levels and list of check functions whose importance you wish to change.
        Typically loaded via json.load from a valid .json file
    checks : list of check functions
        Defaults to all registered checks.
    ignore: list, optional
        Names of functions to skip.
    select: list, optional
        If loading all registered checks, this can be shorthand for selecting only a handful of them.
    importance_threshold : string, optional
        Ignores all tests with an post-configuration assigned importance below this threshold.
        Importance has three levels:

            CRITICAL
                - potentially incorrect data
            BEST_PRACTICE_VIOLATION
                - very suboptimal data representation
            BEST_PRACTICE_SUGGESTION
                - improvable data representation

        The default is the lowest level, BEST_PRACTICE_SUGGESTION.
    """
    if ignore is not None and select is not None:
        raise ValueError("Options 'ignore' and 'select' cannot both be used.")
    if importance_threshold not in Importance:
        raise ValueError(
            f"Indicated importance_threshold ({importance_threshold}) is not a valid importance level! Please choose "
            "from [CRITICAL_IMPORTANCE, BEST_PRACTICE_VIOLATION, BEST_PRACTICE_SUGGESTION]."
        )
    if config is not None:
        validate_config(config=config)
        checks_out = []
        ignore = ignore or []
        for check in checks:
            mapped_check = copy_check(check=check)
            for importance_name, func_names in config.items():
                if check.__name__ in func_names:
                    if importance_name == "SKIP":
                        ignore.append(check.__name__)
                        continue
                    mapped_check.importance = Importance[importance_name]
                    # Output wrappers are apparently parsed at time of wrapping not of time of output return...
                    # Attempting to re-wrap the copied function if the importance level is being adjusted...
                    # From https://github.com/NeurodataWithoutBorders/nwbinspector/issues/302
                    # new_check_wrapper = _copy_function(function=mapped_check.__wrapped__)
                    # new_check_wrapper.importance = Importance[importance_name]
                    # mapped_check.__wrapped__ = new_check_wrapper
            checks_out.append(mapped_check)
    else:
        checks_out = checks
    if select:
        checks_out = [x for x in checks_out if x.__name__ in select]
    elif ignore:
        checks_out = [x for x in checks_out if x.__name__ not in ignore]
    if importance_threshold:
        checks_out = [x for x in checks_out if x.importance.value >= importance_threshold.value]
    return checks_out


@click.command()
@click.argument("path")
@click.option("--modules", help="Modules to import prior to reading the file(s).")
@click.option(
    "--report-file-path",
    default=None,
    help="Save path for the report file.",
    type=click.Path(writable=True),
)
@click.option("--levels", help="Comma-separated names of InspectorMessage attributes to organize by.")
@click.option(
    "--reverse", help="Comma-separated booleans corresponding to reversing the order for each value of 'levels'."
)
@click.option("--overwrite", help="Overwrite an existing report file at the location.", is_flag=True)
@click.option("--ignore", help="Comma-separated names of checks to skip.")
@click.option("--select", help="Comma-separated names of checks to run.")
@click.option(
    "--threshold",
    default="BEST_PRACTICE_SUGGESTION",
    type=click.Choice(["CRITICAL", "BEST_PRACTICE_VIOLATION", "BEST_PRACTICE_SUGGESTION"]),
    help="Ignores tests with an assigned importance below this threshold.",
)
@click.option("--config", help="Name of config or path of config .yaml file that overwrites importance of checks.")
@click.option("--json-file-path", help="Write json output to this location.")
@click.option("--n-jobs", help="Number of jobs to use in parallel.", default=1)
@click.option("--skip-validate", help="Skip the PyNWB validation step.", is_flag=True)
@click.option(
    "--detailed",
    help=(
        "If file_path is the last of 'levels' (the default), identical checks will be aggregated in the display. "
        "Use '--detailed' to see the complete report."
    ),
    is_flag=True,
)
@click.option("--progress-bar", help="Set this flag to False to disable display of the progress bar.")
@click.option(
    "--stream",
    help=(
        "Stream data from the DANDI archive. If the 'path' is a local copy of the target DANDISet, specifying this "
        "flag will still force the data to be streamed instead of using the local copy. To use the local copy, simply "
        "remove this flag. Requires the Read Only S3 (ros3) driver to be installed with h5py."
    ),
    is_flag=True,
)
@click.option(
    "--version-id",
    help=(
        "When 'path' is a six-digit DANDISet ID, this further specifies which version of " "the DANDISet to inspect."
    ),
)
def inspect_all_cli(
    path: str,
    modules: Optional[str] = None,
    report_file_path: str = None,
    levels: str = None,
    reverse: Optional[str] = None,
    overwrite: bool = False,
    ignore: Optional[str] = None,
    select: Optional[str] = None,
    threshold: str = "BEST_PRACTICE_SUGGESTION",
    config: Optional[str] = None,
    json_file_path: Optional[str] = None,
    n_jobs: int = 1,
    skip_validate: bool = False,
    detailed: bool = False,
    progress_bar: Optional[str] = None,
    stream: bool = False,
    version_id: Optional[str] = None,
):
    """
    Run the NWBInspector via the command line.

    path :
    Path to either a local NWBFile, a local folder containing NWBFiles, a link to a dataset on
    DANDI archive (i.e., https://dandiarchive.org/dandiset/{dandiset_id}/{version_id}), or a six-digit Dandiset ID.
    """
    levels = ["importance", "file_path"] if levels is None else levels.split(",")
    reverse = [False] * len(levels) if reverse is None else [strtobool(x) for x in reverse.split(",")]
    progress_bar = strtobool(progress_bar) if progress_bar is not None else True
    if config is not None:
        config = load_config(filepath_or_keyword=config)
    if stream:
        url_path = path if path.startswith("https://") else None
        if url_path:
            dandiset_id, version_id = url_path.split("/")[-2:]
            path = dandiset_id
        assert url_path or re.fullmatch(
            pattern="^[0-9]{6}$", string=path
        ), "'--stream' flag was enabled, but 'path' is neither a full link to the DANDI archive nor a DANDISet ID."
        if Path(path).is_dir():
            warn(
                f"The local DANDISet '{path}' exists, but the '--stream' flag was used. "
                "NWBInspector will use S3 streaming from DANDI. To use local data, remove the '--stream' flag."
            )
    messages = list(
        inspect_all(
            path=path,
            modules=modules,
            ignore=ignore if ignore is None else ignore.split(","),
            select=select if select is None else select.split(","),
            importance_threshold=Importance[threshold],
            config=config,
            n_jobs=n_jobs,
            skip_validate=skip_validate,
            progress_bar=progress_bar,
            stream=stream,
            version_id=version_id,
        )
    )
    if json_file_path is not None:
        if Path(json_file_path).exists() and not overwrite:
            raise FileExistsError(f"The file {json_file_path} already exists! Specify the '-o' flag to overwrite.")
        with open(file=json_file_path, mode="w") as fp:
            json_report = dict(header=get_report_header(), messages=messages)
            json.dump(obj=json_report, fp=fp, cls=InspectorOutputJSONEncoder)
            print(f"{os.linesep*2}Report saved to {str(Path(json_file_path).absolute())}!{os.linesep}")
    formatted_messages = format_messages(messages=messages, levels=levels, reverse=reverse, detailed=detailed)
    print_to_console(formatted_messages=formatted_messages)
    if report_file_path is not None:
        save_report(report_file_path=report_file_path, formatted_messages=formatted_messages, overwrite=overwrite)
        print(f"{os.linesep*2}Report saved to {str(Path(report_file_path).absolute())}!{os.linesep}")


def inspect_all(
    path: PathType,
    modules: OptionalListOfStrings = None,
    config: Optional[dict] = None,
    ignore: OptionalListOfStrings = None,
    select: OptionalListOfStrings = None,
    importance_threshold: Union[str, Importance] = Importance.BEST_PRACTICE_SUGGESTION,
    n_jobs: int = 1,
    skip_validate: bool = False,
    progress_bar: bool = True,
    progress_bar_options: Optional[dict] = None,
    stream: bool = False,
    version_id: Optional[str] = None,
):
    """
    Inspect a local NWBFile or folder of NWBFiles and return suggestions for improvements according to best practices.

    Parameters
    ----------
    path : PathType
        File path to an NWBFile, folder path to iterate over recursively and scan all NWBFiles present, or a
        six-digit identifier of the DANDISet.
    modules : list of strings, optional
        List of external module names to load; examples would be namespace extensions.
        These modules may also contain their own custom checks for their extensions.
    config : dict, optional
        If a dictionary, it must be valid against our JSON configuration schema.
        Can specify a mapping of importance levels and list of check functions whose importance you wish to change.
        Typically loaded via json.load from a valid .json file
    ignore: list of strings, optional
        Names of functions to skip.
    select: list of strings, optional
        Names of functions to pick out of available checks.
    importance_threshold : string or Importance, optional
        Ignores tests with an assigned importance below this threshold.
        Importance has three levels:

            CRITICAL
                - potentially incorrect data
            BEST_PRACTICE_VIOLATION
                - very suboptimal data representation
            BEST_PRACTICE_SUGGESTION
                - improvable data representation

        The default is the lowest level, BEST_PRACTICE_SUGGESTION.
    n_jobs : int
        Number of jobs to use in parallel. Set to -1 to use all available resources.
        This may also be a negative integer x from -2 to -(number_of_cpus - 1) which acts like negative slicing by using
        all available CPUs minus x.
        Set to 1 (also the default) to disable.
    skip_validate : bool, optional
        Skip the PyNWB validation step. This may be desired for older NWBFiles (< schema version v2.10).
        The default is False, which is also recommended.
    progress_bar : bool, optional
        Display a progress bar while scanning NWBFiles.
        Defaults to True.
    progress_bar_options : dict, optional
        Dictionary of keyword arguments to pass directly to tqdm.
    stream : bool, optional
        Stream data from the DANDI archive. If the 'path' is a local copy of the target DANDISet, setting this
        argument to True will force the data to be streamed instead of using the local copy.
        Requires the Read Only S3 (ros3) driver to be installed with h5py.
        Defaults to False.
    version_id : str, optional
        If the path is a DANDISet ID, version_id additionally specifies which version of the dataset to read from.
        Common options are 'draft' or 'published'.
        Defaults to the most recent published version, or if not published then the most recent draft version.
    """
    importance_threshold = (
        Importance[importance_threshold] if isinstance(importance_threshold, str) else importance_threshold
    )
    modules = modules or []
    n_jobs = calculate_number_of_cpu(requested_cpu=n_jobs)
    if progress_bar_options is None:
        progress_bar_options = dict(position=0, leave=False)
        if stream:
            progress_bar_options.update(desc="Inspecting NWBFiles with ROS3...")
        else:
            progress_bar_options.update(desc="Inspecting NWBFiles...")
    if stream:
        assert (
            re.fullmatch(pattern="^[0-9]{6}$", string=str(path)) is not None
        ), "'--stream' flag was enabled, but 'path' is not a DANDISet ID."
        driver = "ros3"
        nwbfiles = get_s3_urls_and_dandi_paths(dandiset_id=path, version_id=version_id, n_jobs=n_jobs)
    else:
        driver = None
        in_path = Path(path)
        if in_path.is_dir():
            nwbfiles = list(in_path.rglob("*.nwb"))
        elif in_path.is_file():
            nwbfiles = [in_path]
        else:
            raise ValueError(f"{in_path} should be a directory or an NWB file.")
    for module in modules:
        importlib.import_module(module)
    # Filtering of checks should apply after external modules are imported, in case those modules have their own checks
    checks = configure_checks(config=config, ignore=ignore, select=select, importance_threshold=importance_threshold)

    # Manual identifier check over all files in the folder path
    identifiers = defaultdict(list)
    for nwbfile_path in nwbfiles:
        with pynwb.NWBHDF5IO(path=nwbfile_path, mode="r", load_namespaces=True, driver=driver) as io:
            nwbfile = robust_s3_read(io.read)
            identifiers[nwbfile.identifier].append(nwbfile_path)
    if len(identifiers) != len(nwbfiles):
        for identifier, nwbfiles_with_identifier in identifiers.items():
            if len(nwbfiles_with_identifier) > 1:
                yield InspectorMessage(
                    message=(
                        f"The identifier '{identifier}' is used across the .nwb files: "
                        f"{natsorted([x.name for x in nwbfiles_with_identifier])}. "
                        "The identifier of any NWBFile should be a completely unique value - "
                        "we recommend using uuid4 to achieve this."
                    ),
                    importance=Importance.CRITICAL,
                    check_function_name="check_unique_identifiers",
                    object_type="NWBFile",
                    object_name="root",
                    location="/",
                    file_path=str(path),
                )

    nwbfiles_iterable = nwbfiles
    if progress_bar:
        nwbfiles_iterable = tqdm(nwbfiles_iterable, **progress_bar_options)
    if n_jobs != 1:
        progress_bar_options.update(total=len(nwbfiles))
        futures = []
        n_jobs = None if n_jobs == -1 else n_jobs  # concurrents uses None instead of -1 for 'auto' mode
        with ProcessPoolExecutor(max_workers=n_jobs) as executor:
            for nwbfile_path in nwbfiles:
                futures.append(
                    executor.submit(
                        _pickle_inspect_nwb,
                        nwbfile_path=nwbfile_path,
                        checks=checks,
                        skip_validate=skip_validate,
                        driver=driver,
                    )
                )
            nwbfiles_iterable = as_completed(futures)
            if progress_bar:
                nwbfiles_iterable = tqdm(nwbfiles_iterable, **progress_bar_options)
            for future in nwbfiles_iterable:
                for message in future.result():
                    if stream:
                        message.file_path = nwbfiles[message.file_path]
                    yield message
    else:
        for nwbfile_path in nwbfiles_iterable:
            for message in inspect_nwb(nwbfile_path=nwbfile_path, checks=checks, driver=driver):
                if stream:
                    message.file_path = nwbfiles[message.file_path]
                yield message


def _pickle_inspect_nwb(
    nwbfile_path: str, checks: list = available_checks, skip_validate: bool = False, driver: Optional[str] = None
):
    """Auxiliary function for inspect_all to run in parallel using the ProcessPoolExecutor."""
    return list(inspect_nwb(nwbfile_path=nwbfile_path, checks=checks, skip_validate=skip_validate, driver=driver))


def inspect_nwb(
    nwbfile_path: FilePathType,
    checks: list = available_checks,
    config: dict = None,
    ignore: OptionalListOfStrings = None,
    select: OptionalListOfStrings = None,
    importance_threshold: Union[str, Importance] = Importance.BEST_PRACTICE_SUGGESTION,
    driver: Optional[str] = None,
    skip_validate: bool = False,
    max_retries: int = 10,
) -> List[InspectorMessage]:
    """
    Inspect a NWBFile object and return suggestions for improvements according to best practices.

    Parameters
    ----------
    nwbfile_path : FilePathType
        Path to the NWBFile.
    checks : list, optional
        list of checks to run
    config : dict
        Dictionary valid against our JSON configuration schema.
        Can specify a mapping of importance levels and list of check functions whose importance you wish to change.
        Typically loaded via json.load from a valid .json file
    ignore: list, optional
        Names of functions to skip.
    select: list, optional
        Names of functions to pick out of available checks.
    importance_threshold : string or Importance, optional
        Ignores tests with an assigned importance below this threshold.
        Importance has three levels:

            CRITICAL
                - potentially incorrect data
            BEST_PRACTICE_VIOLATION
                - very suboptimal data representation
            BEST_PRACTICE_SUGGESTION
                - improvable data representation

        The default is the lowest level, BEST_PRACTICE_SUGGESTION.
    driver: str, optional
        Forwarded to h5py.File(). Set to "ros3" for reading from s3 url.
    skip_validate : bool
        Skip the PyNWB validation step. This may be desired for older NWBFiles (< schema version v2.10).
        The default is False, which is also recommended.
    max_retries : int, optional
        When using the ros3 driver to stream data from an s3 path, occasional curl issues can result.
        AWS suggests using iterative retry with an exponential backoff of 0.1 * 2^retries.
        This sets a hard bound on the number of times to attempt to retry the collection of messages.
        Defaults to 10 (corresponds to 102.4s maximum delay on final attempt).
    """
    importance_threshold = (
        Importance[importance_threshold] if isinstance(importance_threshold, str) else importance_threshold
    )
    if any(x is not None for x in [config, ignore, select, importance_threshold]):
        checks = configure_checks(
            checks=checks, config=config, ignore=ignore, select=select, importance_threshold=importance_threshold
        )
    nwbfile_path = str(nwbfile_path)
    filterwarnings(action="ignore", message="No cached namespaces found in .*")
    filterwarnings(action="ignore", message="Ignoring cached namespace .*")

    with pynwb.NWBHDF5IO(path=nwbfile_path, mode="r", load_namespaces=True, driver=driver) as io:
        if not skip_validate:
            validation_errors = pynwb.validate(io=io)
            for validation_error in validation_errors:
                yield InspectorMessage(
                    message=validation_error.reason,
                    importance=Importance.PYNWB_VALIDATION,
                    check_function_name=validation_error.name,
                    location=validation_error.location,
                    file_path=nwbfile_path,
                )

        try:
            nwbfile = robust_s3_read(command=io.read, max_retries=max_retries)
            for inspector_message in run_checks(nwbfile=nwbfile, checks=checks):
                inspector_message.file_path = nwbfile_path
                yield inspector_message
        except Exception as ex:
            yield InspectorMessage(
                message=traceback.format_exc(),
                importance=Importance.ERROR,
                check_function_name=f"During io.read() - {type(ex)}: {str(ex)}",
                file_path=nwbfile_path,
            )


def run_checks(nwbfile: pynwb.NWBFile, checks: list):
    """
    Run checks on an open NWBFile object.

    Parameters
    ----------
    nwbfile : NWBFile
    checks : list
    """
    for check_function in checks:
        for nwbfile_object in nwbfile.objects.values():
            if check_function.neurodata_type is None or issubclass(type(nwbfile_object), check_function.neurodata_type):
                try:
                    output = robust_s3_read(command=check_function, command_args=[nwbfile_object])
                # if an individual check fails, include it in the report and continue with the inspection
                except Exception:
                    output = InspectorMessage(
                        message=traceback.format_exc(),
                        importance=Importance.ERROR,
                        check_function_name=check_function.__name__,
                    )
                if isinstance(output, InspectorMessage) and output.importance != Importance.ERROR:
                    # temporary solution to https://github.com/dandi/dandi-cli/issues/1031
                    output.importance = check_function.importance
                    yield output
                elif output is not None:
                    for x in output:
                        x.importance = check_function.importance
                        yield x


if __name__ == "__main__":
    inspect_all_cli()



import os
from packaging import version

from hdmf.testing import TestCase

from nwbinspector import Importance
from nwbinspector.utils import (
    format_byte_size,
    is_regular_series,
    is_dict_in_string,
    get_package_version,
    calculate_number_of_cpu,
)


def test_format_byte_size():
    assert format_byte_size(byte_size=12345) == "12.35KB"


def test_format_byte_size_in_binary():
    assert format_byte_size(byte_size=12345, units="binary") == "12.06KiB"


class TestFormatByteException(TestCase):
    def test_format_byte_size_units_exception(self):
        with self.assertRaisesWith(
            exc_type=ValueError,
            exc_msg="'units' argument must be either 'SI' (for orders of 1000) or 'binary' (for orders of 1024).",
        ):
            format_byte_size(byte_size=12345, units="test")


def test_is_regular_series():
    assert is_regular_series(series=[1, 2, 3])
    assert not is_regular_series(series=[1, 2, 4])


def test_is_dict_in_string_false_1():
    string = "not a dict"
    assert is_dict_in_string(string=string) is False


def test_is_dict_in_string_false_2():
    string = "not a dict, {but also fancy format text!}"
    assert is_dict_in_string(string=string) is False


def test_is_dict_in_string_false_3():
    string = "[not] a dict, {[but] also} fancy format text!"
    assert is_dict_in_string(string=string) is False


def test_is_dict_in_string_true_1():
    string = str(dict(a=1))
    assert is_dict_in_string(string=string) is True


def test_is_dict_in_string_true_2():
    string = str([dict(a=1), dict(b=2)])
    assert is_dict_in_string(string=string) is True


def test_is_dict_in_string_true_3():
    string = str(dict(a=dict(b=2)))
    assert is_dict_in_string(string=string) is True


def test_is_dict_in_string_true_4():
    string = "some text: {'then': 'a dict'}"
    assert is_dict_in_string(string=string) is True


def test_is_dict_in_string_true_5():
    string = "{'then': 'a dict'} more text"
    assert is_dict_in_string(string=string) is True


def test_is_dict_in_string_true_6():
    """Not a JSON encodable object."""
    string = str({1.2: Importance.CRITICAL})
    assert is_dict_in_string(string=string) is True


def test_is_dict_in_string_true_7():
    """
    A more aggressive demonstration of the general dictionary regex.

    But it is technically possible to achieve via `str({custom_object_1: custom_object_2})` if 'custom_object_1' is
    hashable and both custom objects have manual `__repr__` that do not include apostrophe's within their return.

    Example
    -------
    from dataclasses import dataclass

    @dataclass
    class Test():
        prop = 1

        def __repr__(self):
            return "This is a test"

    str({1: Test()})
    """
    string = "example, {this is not a dict: but it sure looks like one}!"
    assert is_dict_in_string(string=string) is True


def test_get_package_version_type():
    assert isinstance(get_package_version("hdmf"), version.Version)


def test_get_package_version_value():
    assert get_package_version("hdmf") >= version.parse("3.1.1")  # minimum supported PyNWB version


class TestCalulcateNumberOfCPU(TestCase):
    total_cpu = os.cpu_count()

    def test_request_more_than_available_assert(self):
        requested_cpu = 2500
        with self.assertRaisesWith(
            exc_type=AssertionError,
            exc_msg=f"Requested more CPUs ({requested_cpu}) than are available ({self.total_cpu})!",
        ):
            calculate_number_of_cpu(requested_cpu=requested_cpu)

    def test_request_fewer_than_available_assert(self):
        requested_cpu = -2500
        with self.assertRaisesWith(
            exc_type=AssertionError,
            exc_msg=f"Requested fewer CPUs ({requested_cpu}) than are available ({self.total_cpu})!",
        ):
            calculate_number_of_cpu(requested_cpu=requested_cpu)

    def test_calculate_number_of_cpu_negative_value(self):
        requested_cpu = -1  # CI only has 2 jobs available
        assert calculate_number_of_cpu(requested_cpu=requested_cpu) == requested_cpu % self.total_cpu



"""Helper functions for internal use across the testing suite."""
import os
import json
from distutils.util import strtobool
from pathlib import Path
from typing import Tuple, Optional

from packaging.version import Version
from pynwb import NWBHDF5IO
from pynwb.image import ImageSeries

from .tools import check_streaming_enabled, make_minimal_nwbfile
from .utils import is_module_installed, get_package_version


def check_streaming_tests_enabled() -> Tuple[bool, Optional[str]]:
    """
    General purpose helper for determining if the testing environment can support S3 DANDI streaming.

    Returns the boolean status of the check and, if False, provides a string reason for the failure for the user to
    utilize as they please (raise an error or warning with that message, print it, or ignore it).
    """
    failure_reason = ""

    environment_skip_flag = os.environ.get("NWBI_SKIP_NETWORK_TESTS", "")
    environment_skip_flag_bool = (
        strtobool(os.environ.get("NWBI_SKIP_NETWORK_TESTS", "")) if environment_skip_flag != "" else False
    )
    if environment_skip_flag_bool:
        failure_reason += "Environmental variable set to skip network tets."

    streaming_enabled, streaming_failure_reason = check_streaming_enabled()
    if not streaming_enabled:
        failure_reason += streaming_failure_reason

    have_dandi = is_module_installed("dandi")
    if not have_dandi:
        failure_reason += "The DANDI package is not installed on the system."

    failure_reason = None if failure_reason == "" else failure_reason
    return streaming_enabled and not environment_skip_flag_bool and have_dandi, failure_reason


def load_testing_config() -> dict:
    """Helper function for loading the testing configuration file as a dictionary."""
    test_config_file_path = Path(__file__).parent.parent.parent / "tests" / "testing_config.json"

    # This error would only occur if someone installed a previous version
    # directly from GitHub and then updated the branch/commit in-place
    if not test_config_file_path.exists():  # pragma: no cover
        raise FileNotFoundError(
            "The testing configuration file not found at the location '{test_config_file_path}'! "
            "Please try reinstalling the package."
        )

    with open(file=test_config_file_path) as file:
        test_config = json.load(file)

    return test_config


def update_testing_config(key: str, value):
    """Update a key/value pair in the testing configuration file through the API."""
    test_config_file_path = Path(__file__).parent.parent.parent / "tests" / "testing_config.json"

    testing_config = load_testing_config()

    if key not in testing_config:
        raise KeyError("Updating the testing configuration file via the API is only possible for the pre-defined keys!")
    testing_config[key] = value

    with open(file=test_config_file_path, mode="w") as file:
        json.dump(testing_config, file)


def generate_testing_files():  # pragma: no cover
    """Generate a local copy of the NWB files required for all tests."""
    generate_image_series_testing_files()


def generate_image_series_testing_files():  # pragma: no cover
    """Generate a local copy of the NWB files required for the image series tests."""
    assert get_package_version(name="pynwb") == Version("2.1.0"), "Generating the testing files requires PyNWB v2.1.0!"

    testing_config = load_testing_config()

    local_path = Path(testing_config["LOCAL_PATH"])
    local_path.mkdir(exist_ok=True, parents=True)

    movie_1_file_path = local_path / "temp_movie_1.mov"
    nested_folder = local_path / "nested_folder"
    nested_folder.mkdir(exist_ok=True)
    movie_2_file_path = nested_folder / "temp_movie_2.avi"
    for file_path in [movie_1_file_path, movie_2_file_path]:
        with open(file=file_path, mode="w") as file:
            file.write("Not a movie file, but at least it exists.")
    nwbfile = make_minimal_nwbfile()
    nwbfile.add_acquisition(
        ImageSeries(
            name="TestImageSeriesGoodExternalPaths",
            rate=1.0,
            external_file=[
                "/".join([".", movie_1_file_path.name]),
                "/".join([".", nested_folder.name, movie_2_file_path.name]),
            ],
        )
    )
    nwbfile.add_acquisition(
        ImageSeries(
            name="TestImageSeriesExternalPathDoesNotExist",
            rate=1.0,
            external_file=["madeup_file.mp4"],
        )
    )
    absolute_file_path = str(Path("madeup_file.mp4").absolute())
    nwbfile.add_acquisition(
        ImageSeries(name="TestImageSeriesExternalPathIsNotRelative", rate=1.0, external_file=[absolute_file_path])
    )
    with NWBHDF5IO(path=local_path / "image_series_testing_file.nwb", mode="w") as io:
        io.write(nwbfile)



"""Check functions that can apply to any descendant of TimeSeries."""
import numpy as np

from pynwb import TimeSeries

# from ..tools import all_of_type
from ..register_checks import register_check, Importance, Severity, InspectorMessage
from ..utils import is_regular_series, is_ascending_series


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=TimeSeries)
def check_regular_timestamps(
    time_series: TimeSeries, time_tolerance_decimals: int = 9, gb_severity_threshold: float = 1.0
):
    """If the TimeSeries uses timestamps, check if they are regular (i.e., they have a constant rate)."""
    if (
        time_series.timestamps is not None
        and len(time_series.timestamps) > 2
        and is_regular_series(series=time_series.timestamps, tolerance_decimals=time_tolerance_decimals)
    ):
        timestamps = np.array(time_series.timestamps)
        if timestamps.size * timestamps.dtype.itemsize > gb_severity_threshold * 1e9:
            severity = Severity.HIGH
        else:
            severity = Severity.LOW
        return InspectorMessage(
            severity=severity,
            message=(
                "TimeSeries appears to have a constant sampling rate. "
                f"Consider specifying starting_time={time_series.timestamps[0]} "
                f"and rate={time_series.timestamps[1] - time_series.timestamps[0]} instead of timestamps."
            ),
        )


@register_check(importance=Importance.CRITICAL, neurodata_type=TimeSeries)
def check_data_orientation(time_series: TimeSeries):
    """If the TimeSeries has data, check if the longest axis (almost always time) is also the zero-axis."""
    if time_series.data is not None and any(np.array(time_series.data.shape[1:]) > time_series.data.shape[0]):
        return InspectorMessage(
            message=(
                "Data may be in the wrong orientation. "
                "Time should be in the first dimension, and is usually the longest dimension. "
                "Here, another dimension is longer."
            ),
        )


@register_check(importance=Importance.CRITICAL, neurodata_type=TimeSeries)
def check_timestamps_match_first_dimension(time_series: TimeSeries):
    """
    If the TimeSeries has timestamps, check if their length is the same as the zero-axis of data.

    Best Practice: :ref:`best_practice_data_orientation`
    """
    if (
        time_series.data is not None
        and time_series.timestamps is not None
        and np.array(time_series.data).shape[:1] != np.array(time_series.timestamps).shape
    ):
        return InspectorMessage(
            message="The length of the first dimension of data does not match the length of timestamps.",
        )


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=TimeSeries)
def check_timestamps_ascending(time_series: TimeSeries, nelems=200):
    """Check that the values in the timestamps array are strictly increasing."""
    if time_series.timestamps is not None and not is_ascending_series(time_series.timestamps, nelems=nelems):
        return InspectorMessage(f"{time_series.name} timestamps are not ascending.")


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=TimeSeries)
def check_missing_unit(time_series: TimeSeries):
    """
    Check if the TimeSeries.unit field is empty.

    Best Practice: :ref:`best_practice_unit_of_measurement`
    """
    if not time_series.unit:
        return InspectorMessage(
            message="Missing text for attribute 'unit'. Please specify the scientific unit of the 'data'."
        )


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=TimeSeries)
def check_resolution(time_series: TimeSeries):
    """Check the resolution value of a TimeSeries for proper format (-1.0 or NaN for unknown)."""
    if time_series.resolution is None or time_series.resolution == -1.0:
        return
    if time_series.resolution <= 0:
        return InspectorMessage(
            message=f"'resolution' should use -1.0 or NaN for unknown instead of {time_series.resolution}."
        )



from datetime import datetime
from unittest import TestCase
from uuid import uuid4

import numpy as np
from pynwb import NWBFile
from pynwb.ecephys import ElectricalSeries
from pynwb.misc import Units
from hdmf.common.table import DynamicTableRegion, DynamicTable

from nwbinspector import (
    InspectorMessage,
    Importance,
    check_negative_spike_times,
    check_electrical_series_dims,
    check_electrical_series_reference_electrodes_table,
    check_spike_times_not_in_unobserved_interval,
)


def test_check_negative_spike_times_all_positive():
    units_table = Units()
    units_table.add_unit(spike_times=[0.0, 0.1])
    units_table.add_unit(spike_times=[1.0])
    assert check_negative_spike_times(units_table=units_table) is None


def test_check_negative_spike_times_empty():
    units_table = Units()
    assert check_negative_spike_times(units_table=units_table) is None


def test_check_negative_spike_times_some_negative():
    units_table = Units()
    units_table.add_unit(spike_times=[0.0, 0.1])
    units_table.add_unit(spike_times=[-1.0])
    assert check_negative_spike_times(units_table=units_table) == InspectorMessage(
        message=(
            "This Units table contains negative spike times. Time should generally be aligned to the earliest time "
            "reference in the NWBFile."
        ),
        importance=Importance.BEST_PRACTICE_VIOLATION,
        check_function_name="check_negative_spike_times",
        object_type="Units",
        object_name="Units",
        location="/",
    )


class TestCheckElectricalSeries(TestCase):
    def setUp(self):

        nwbfile = NWBFile(
            session_description="", identifier=str(uuid4()), session_start_time=datetime.now().astimezone()
        )

        device = nwbfile.create_device(name="dev")
        group = nwbfile.create_electrode_group(
            name="electrode_group", description="desc", location="loc", device=device
        )

        for _ in range(5):
            nwbfile.add_electrode(
                x=3.0,
                y=3.0,
                z=3.0,
                imp=-1.0,
                location="unknown",
                filtering="unknown",
                group=group,
            )
        self.nwbfile = nwbfile

    def test_check_electrical_series_wrong_dims(self):

        electrodes = self.nwbfile.create_electrode_table_region(region=[1, 2, 3], description="three elecs")

        electrical_series = ElectricalSeries(
            name="elec_series",
            description="desc",
            data=np.zeros((100, 10)),
            electrodes=electrodes,
            rate=30.0,
        )

        self.nwbfile.add_acquisition(electrical_series)

        assert check_electrical_series_dims(self.nwbfile.acquisition["elec_series"]) == InspectorMessage(
            message=(
                "The second dimension of data does not match the length of electrodes. Your data may be transposed."
            ),
            importance=Importance.CRITICAL,
            check_function_name="check_electrical_series_dims",
            object_type="ElectricalSeries",
            object_name="elec_series",
        )

    def test_check_electrical_series_flipped(self):

        electrodes = self.nwbfile.create_electrode_table_region(region=[0, 1, 2, 3, 4], description="all")

        electrical_series = ElectricalSeries(
            name="elec_series",
            description="desc",
            data=np.zeros((5, 100)),
            electrodes=electrodes,
            rate=30.0,
        )

        self.nwbfile.add_acquisition(electrical_series)

        assert check_electrical_series_dims(self.nwbfile.acquisition["elec_series"]) == InspectorMessage(
            message=(
                "The second dimension of data does not match the length of electrodes, but instead the first does. "
                "Data is oriented incorrectly and should be transposed."
            ),
            importance=Importance.CRITICAL,
            check_function_name="check_electrical_series_dims",
            object_type="ElectricalSeries",
            object_name="elec_series",
        )

    def test_pass(self):

        electrodes = self.nwbfile.create_electrode_table_region(region=[0, 1, 2, 3, 4], description="all")

        electrical_series = ElectricalSeries(
            name="elec_series",
            description="desc",
            data=np.zeros((100, 5)),
            electrodes=electrodes,
            rate=30.0,
        )

        self.nwbfile.add_acquisition(electrical_series)

        assert check_electrical_series_dims(electrical_series) is None
        assert check_electrical_series_reference_electrodes_table(electrical_series) is None

    def test_trigger_check_electrical_series_reference_electrodes_table(self):

        dyn_tab = DynamicTable("name", "desc")
        dyn_tab.add_column("name", "desc")
        for i in range(5):
            dyn_tab.add_row(name=1)
        dynamic_table_region = DynamicTableRegion(
            name="electrodes", description="I am wrong", data=[0, 1, 2, 3, 4], table=dyn_tab
        )

        electrical_series = ElectricalSeries(
            name="elec_series",
            description="desc",
            data=np.zeros((100, 5)),
            electrodes=dynamic_table_region,
            rate=30.0,
        )

        assert (
            check_electrical_series_reference_electrodes_table(electrical_series).message
            == "electrodes does not  reference an electrodes table."
        )


def test_check_spike_times_not_in_unobserved_interval_pass():
    units_table = Units(name="TestUnits")
    units_table.add_unit(spike_times=[1, 2, 3], obs_intervals=[[0, 2.5], [2.7, 3.5]])
    assert check_spike_times_not_in_unobserved_interval(units_table=units_table) is None


def test_check_spike_times_not_in_unobserved_interval_pass_no_intervals():
    units_table = Units(name="TestUnits")
    units_table.add_unit(spike_times=[1, 2, 3])
    assert check_spike_times_not_in_unobserved_interval(units_table=units_table) is None


def test_check_spike_times_not_in_unobserved_interval_1():
    units_table = Units(name="TestUnits")
    units_table.add_unit(spike_times=[1, 2, 3], obs_intervals=[[0, 2.5], [3.5, 4]])
    assert check_spike_times_not_in_unobserved_interval(units_table=units_table) == InspectorMessage(
        message=(
            "This Units table contains spike times that occur during periods of time not labeled as being "
            "observed intervals."
        ),
        importance=Importance.BEST_PRACTICE_VIOLATION,
        check_function_name="check_spike_times_not_in_unobserved_interval",
        object_type="Units",
        object_name="TestUnits",
        location="/",
    )


def test_check_spike_times_not_in_unobserved_interval_2():
    units_table = Units(name="TestUnits")
    units_table.add_unit(spike_times=[1, 2, 3, 4, 5, 6], obs_intervals=[[0, 2.5], [3.5, 7]])
    assert check_spike_times_not_in_unobserved_interval(units_table=units_table) == InspectorMessage(
        message=(
            "This Units table contains spike times that occur during periods of time not labeled as being "
            "observed intervals."
        ),
        importance=Importance.BEST_PRACTICE_VIOLATION,
        check_function_name="check_spike_times_not_in_unobserved_interval",
        object_type="Units",
        object_name="TestUnits",
        location="/",
    )


def test_check_spike_times_not_in_unobserved_interval_multiple_units():
    units_table = Units(name="TestUnits")
    units_table.add_unit(spike_times=[1, 2, 3, 4, 5, 6], obs_intervals=[[0, 3.2], [3.5, 7]])
    units_table.add_unit(spike_times=[6.5, 12, 13, 14, 15, 16], obs_intervals=[[10, 15.2], [15.8, 17]])
    assert check_spike_times_not_in_unobserved_interval(units_table=units_table) == InspectorMessage(
        message=(
            "This Units table contains spike times that occur during periods of time not labeled as being "
            "observed intervals."
        ),
        importance=Importance.BEST_PRACTICE_VIOLATION,
        check_function_name="check_spike_times_not_in_unobserved_interval",
        object_type="Units",
        object_name="TestUnits",
        location="/",
    )



from pynwb.ogen import OptogeneticSeries, OptogeneticStimulusSite

from ..register_checks import register_check, Importance, InspectorMessage


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=OptogeneticStimulusSite)
def check_optogenetic_stimulus_site_has_optogenetic_series(ogen_site: OptogeneticStimulusSite):
    """Check if an optogenetic stimulus site has an optogenetic series linked to it."""
    nwbfile = ogen_site.get_ancestor("NWBFile")
    for obj in nwbfile.objects.values():
        if isinstance(obj, OptogeneticSeries):
            if obj.site == ogen_site:
                return
    return InspectorMessage(message="OptogeneticStimulusSite is not referenced by any OptogeneticStimulusSite.")



import unittest
from pathlib import Path

import numpy as np
from pynwb import NWBHDF5IO
from pynwb.image import ImageSeries

from nwbinspector import (
    InspectorMessage,
    Importance,
    check_image_series_external_file_valid,
    check_image_series_external_file_relative,
    check_image_series_data_size,
)
from nwbinspector.testing import load_testing_config

try:
    testing_config = load_testing_config()
    testing_file = Path(testing_config["LOCAL_PATH"]) / "image_series_testing_file.nwb"
    NO_CONFIG = False  # Depending on the method of installation, a config may not have generated
except FileNotFoundError:
    testing_file = "Not found"
    NO_CONFIG = True


@unittest.skipIf(
    NO_CONFIG or not testing_file.exists(),
    reason=f"The ImageSeries unit tests were skipped because the required file ({testing_file}) was not found!",
)
class TestExternalFileValid(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        testing_config = load_testing_config()
        cls.testing_file = Path(testing_config["LOCAL_PATH"]) / "image_series_testing_file.nwb"

    def setUp(self):
        self.io = NWBHDF5IO(path=self.testing_file, mode="r")
        self.nwbfile = self.io.read()

    def tearDown(self):
        self.io.close()

    def test_check_image_series_external_file_valid_pass(self):
        assert (
            check_image_series_external_file_valid(
                image_series=self.nwbfile.acquisition["TestImageSeriesGoodExternalPaths"]
            )
            is None
        )

    def test_check_image_series_external_file_valid_bytestring_pass(self):
        """Can't use the NWB file since the call to io.write() decodes the bytes with modern versions of h5py."""
        good_external_path = Path(self.nwbfile.acquisition["TestImageSeriesGoodExternalPaths"].external_file[0])
        image_series = ImageSeries(
            name="TestImageSeries",
            rate=1.0,
            external_file=[bytes("/".join([".", good_external_path.name]), "utf-8")],
        )
        assert check_image_series_external_file_relative(image_series=image_series) is None

    def test_check_image_series_external_file_valid(self):
        with NWBHDF5IO(path=self.testing_file, mode="r") as io:
            nwbfile = io.read()
            image_series = nwbfile.acquisition["TestImageSeriesExternalPathDoesNotExist"]

            assert check_image_series_external_file_valid(image_series=image_series)[0] == InspectorMessage(
                message=(
                    "The external file 'madeup_file.mp4' does not exist. Please confirm the relative location to the"
                    " NWBFile."
                ),
                importance=Importance.CRITICAL,
                check_function_name="check_image_series_external_file_valid",
                object_type="ImageSeries",
                object_name="TestImageSeriesExternalPathDoesNotExist",
                location="/acquisition/TestImageSeriesExternalPathDoesNotExist",
            )

    def test_check_image_series_external_file_relative_pass(self):
        with NWBHDF5IO(path=self.testing_file, mode="r") as io:
            nwbfile = io.read()

            assert (
                check_image_series_external_file_relative(
                    image_series=nwbfile.acquisition["TestImageSeriesGoodExternalPaths"]
                )
                is None
            )

    def test_check_image_series_external_file_relative_trigger(self):
        with NWBHDF5IO(path=self.testing_file, mode="r") as io:
            nwbfile = io.read()
            image_series = nwbfile.acquisition["TestImageSeriesExternalPathIsNotRelative"]

            assert check_image_series_external_file_relative(image_series=image_series)[0] == InspectorMessage(
                message=(
                    f"The external file '{image_series.external_file[0]}' is not a relative path. "
                    "Please adjust the absolute path to be relative to the location of the NWBFile."
                ),
                importance=Importance.BEST_PRACTICE_VIOLATION,
                check_function_name="check_image_series_external_file_relative",
                object_type="ImageSeries",
                object_name="TestImageSeriesExternalPathIsNotRelative",
                location="/acquisition/TestImageSeriesExternalPathIsNotRelative",
            )


def test_check_image_series_external_file_valid_pass_non_external():
    image_series = ImageSeries(name="TestImageSeries", rate=1.0, data=np.zeros(shape=(3, 3, 3, 3)), unit="TestUnit")

    assert check_image_series_external_file_valid(image_series=image_series) is None


def test_check_small_image_series_stored_internally():
    gb_size = 0.010  # 10 MB
    frame_length = 10
    total_elements = int(gb_size * 1e9 / np.dtype("float").itemsize) // (frame_length * frame_length)
    data = np.zeros(shape=(total_elements, frame_length, frame_length, 1))
    image_series = ImageSeries(name="ImageSeriesLarge", rate=1.0, data=data, unit="TestUnit")

    assert check_image_series_data_size(image_series=image_series) is None


def test_check_large_image_series_stored_internally():
    gb_size = 0.010  # 10 MB
    frame_length = 10
    total_elements = int(gb_size * 1e9 / np.dtype("float").itemsize) // (frame_length * frame_length)
    data = np.zeros(shape=(total_elements, frame_length, frame_length, 1))
    image_series = ImageSeries(name="ImageSeriesLarge", rate=1.0, data=data, unit="TestUnit")
    gb_lower_bound = gb_size * 0.9
    inspector_message = check_image_series_data_size(image_series=image_series, gb_lower_bound=gb_lower_bound)

    expected_message = InspectorMessage(
        importance=Importance.BEST_PRACTICE_VIOLATION,
        message=f"ImageSeries {image_series.name} is too large. Use external mode for storage",
        check_function_name="check_image_series_data_size",
        object_type="ImageSeries",
        object_name="ImageSeriesLarge",
        location="/",
    )

    assert inspector_message == expected_message



from datetime import datetime
from unittest import TestCase
from uuid import uuid4

import numpy as np
from pynwb import NWBFile
from pynwb.device import Device
from pynwb.ophys import (
    OpticalChannel,
    ImageSegmentation,
    RoiResponseSeries,
    ImagingPlane,
    PlaneSegmentation,
    TwoPhotonSeries,
)
from hdmf.common.table import DynamicTableRegion, DynamicTable

from nwbinspector import (
    InspectorMessage,
    Importance,
    check_roi_response_series_dims,
    check_roi_response_series_link_to_plane_segmentation,
    check_excitation_lambda_in_nm,
    check_emission_lambda_in_nm,
    check_plane_segmentation_image_mask_shape_against_ref_images,
)


class TestCheckRoiResponseSeries(TestCase):
    def setUp(self):

        nwbfile = NWBFile(
            session_description="", identifier=str(uuid4()), session_start_time=datetime.now().astimezone()
        )

        device = nwbfile.create_device(
            name="Microscope", description="My two-photon microscope", manufacturer="The best microscope manufacturer"
        )
        optical_channel = OpticalChannel(name="OpticalChannel", description="an optical channel", emission_lambda=500.0)
        imaging_plane = nwbfile.create_imaging_plane(
            name="ImagingPlane",
            optical_channel=optical_channel,
            imaging_rate=30.0,
            description="a very interesting part of the brain",
            device=device,
            excitation_lambda=600.0,
            indicator="GFP",
            location="V1",
            grid_spacing=[0.01, 0.01],
            grid_spacing_unit="meters",
            origin_coords=[1.0, 2.0, 3.0],
            origin_coords_unit="meters",
        )

        img_seg = ImageSegmentation()

        self.plane_segmentation = img_seg.create_plane_segmentation(
            name="PlaneSegmentation",
            description="output from segmenting my favorite imaging plane",
            imaging_plane=imaging_plane,
        )

        self.ophys_module = nwbfile.create_processing_module(
            name="ophys",
            description="optical physiology processed data",
        )

        self.ophys_module.add(img_seg)

        for _ in range(10):
            image_mask = np.zeros((100, 100))
            self.plane_segmentation.add_roi(image_mask=image_mask)
        self.nwbfile = nwbfile

    def test_check_flipped_dims(self):

        rt_region = self.plane_segmentation.create_roi_table_region(
            region=[0, 1, 2, 3, 4],
            description="the first of two ROIs",
        )

        roi_resp_series = RoiResponseSeries(
            name="RoiResponseSeries",
            data=np.ones((5, 40)),  # 50 samples, 2 ROIs
            rois=rt_region,
            unit="n.a.",
            rate=30.0,
        )

        self.ophys_module.add(roi_resp_series)

        assert check_roi_response_series_dims(roi_resp_series) == InspectorMessage(
            message=(
                "The second dimension of data does not match the length of rois, "
                "but instead the first does. Data is oriented incorrectly and should be transposed."
            ),
            importance=Importance.CRITICAL,
            check_function_name="check_roi_response_series_dims",
            object_type="RoiResponseSeries",
            object_name="RoiResponseSeries",
        )

    def test_check_wrong_dims(self):

        rt_region = self.plane_segmentation.create_roi_table_region(
            region=[0, 1, 2, 3, 4],
            description="the first of two ROIs",
        )

        roi_resp_series = RoiResponseSeries(
            name="RoiResponseSeries",
            data=np.ones((10, 40)),  # 50 samples, 2 ROIs
            rois=rt_region,
            unit="n.a.",
            rate=30.0,
        )

        self.ophys_module.add(roi_resp_series)

        assert check_roi_response_series_dims(roi_resp_series) == InspectorMessage(
            message="The second dimension of data does not match the length of rois. Your " "data may be transposed.",
            importance=Importance.CRITICAL,
            check_function_name="check_roi_response_series_dims",
            object_type="RoiResponseSeries",
            object_name="RoiResponseSeries",
        )

    def test_pass_check_roi_response_series_dims(self):
        rt_region = self.plane_segmentation.create_roi_table_region(
            region=[0, 1, 2, 3, 4],
            description="the first of two ROIs",
        )

        roi_resp_series = RoiResponseSeries(
            name="RoiResponseSeries",
            data=np.ones((40, 5)),  # 50 samples, 2 ROIs
            rois=rt_region,
            unit="n.a.",
            rate=30.0,
        )

        assert check_roi_response_series_dims(roi_resp_series) is None

    def test_check_roi_response_series_link_to_plane_segmentation(self):
        dt = DynamicTable(name="name", description="desc")
        dt.add_column("a", "desc")
        for _ in range(5):
            dt.add_row(a=1)
        dtr = DynamicTableRegion(name="n", description="desc", data=[0, 1, 2, 3, 4], table=dt)
        roi_resp_series = RoiResponseSeries(
            name="RoiResponseSeries",
            data=np.ones((40, 5)),  # 50 samples, 2 ROIs
            rois=dtr,
            unit="n.a.",
            rate=30.0,
        )

        self.ophys_module.add(roi_resp_series)

        assert check_roi_response_series_link_to_plane_segmentation(roi_resp_series) == InspectorMessage(
            message="rois field does not point to a PlaneSegmentation table.",
            importance=Importance.BEST_PRACTICE_VIOLATION,
            check_function_name="check_roi_response_series_link_to_plane_segmentation",
            object_type="RoiResponseSeries",
            object_name="RoiResponseSeries",
        )

    def test_pass_check_roi_response_series_link_to_plane_segmentation(self):

        rt_region = self.plane_segmentation.create_roi_table_region(
            region=[0, 1, 2, 3, 4],
            description="the first of two ROIs",
        )

        roi_resp_series = RoiResponseSeries(
            name="RoiResponseSeries",
            data=np.ones((40, 5)),  # 50 samples, 2 ROIs
            rois=rt_region,
            unit="n.a.",
            rate=30.0,
        )

        assert check_roi_response_series_link_to_plane_segmentation(roi_resp_series) is None


def test_check_excitation_lambda_in_nm():

    device = Device(
        name="Microscope", description="My two-photon microscope", manufacturer="The best microscope manufacturer"
    )
    optical_channel = OpticalChannel(name="OpticalChannel", description="an optical channel", emission_lambda=500.0)
    imaging_plane = ImagingPlane(
        name="ImagingPlane",
        optical_channel=optical_channel,
        imaging_rate=30.0,
        description="a very interesting part of the brain",
        device=device,
        excitation_lambda=1.0,
        indicator="GFP",
        location="V1",
        grid_spacing=[0.01, 0.01],
        grid_spacing_unit="meters",
        origin_coords=[1.0, 2.0, 3.0],
        origin_coords_unit="meters",
    )

    assert check_excitation_lambda_in_nm(imaging_plane).message == "excitation lambda of 1.0 should be in units of nm."


def test_pass_check_excitation_lambda_in_nm():
    device = Device(
        name="Microscope", description="My two-photon microscope", manufacturer="The best microscope manufacturer"
    )
    optical_channel = OpticalChannel(name="OpticalChannel", description="an optical channel", emission_lambda=500.0)
    imaging_plane = ImagingPlane(
        name="ImagingPlane",
        optical_channel=optical_channel,
        imaging_rate=30.0,
        description="a very interesting part of the brain",
        device=device,
        excitation_lambda=300.0,
        indicator="GFP",
        location="V1",
        grid_spacing=[0.01, 0.01],
        grid_spacing_unit="meters",
        origin_coords=[1.0, 2.0, 3.0],
        origin_coords_unit="meters",
    )

    assert check_excitation_lambda_in_nm(imaging_plane) is None


def test_check_emission_lambda_in_nm():
    optical_channel = OpticalChannel(name="OpticalChannel", description="an optical channel", emission_lambda=5.0)
    assert check_emission_lambda_in_nm(optical_channel).message == "emission lambda of 5.0 should be in units of nm."


def test_pass_check_emission_lambda_in_nm():
    optical_channel = OpticalChannel(name="OpticalChannel", description="an optical channel", emission_lambda=500.0)
    assert check_emission_lambda_in_nm(optical_channel) is None


def test_pass_check_plane_segmentation_image_mask_dims_against_imageseries():

    device = Device(
        name="Microscope", description="My two-photon microscope", manufacturer="The best microscope manufacturer"
    )
    optical_channel = OpticalChannel(name="OpticalChannel", description="an optical channel", emission_lambda=500.0)
    imaging_plane = ImagingPlane(
        name="ImagingPlane",
        optical_channel=optical_channel,
        imaging_rate=30.0,
        description="a very interesting part of the brain",
        device=device,
        excitation_lambda=300.0,
        indicator="GFP",
        location="V1",
        grid_spacing=[0.01, 0.01],
        grid_spacing_unit="meters",
        origin_coords=[1.0, 2.0, 3.0],
        origin_coords_unit="meters",
    )

    two_photon_series = TwoPhotonSeries(
        name="TwoPhotonSeries",
        imaging_plane=imaging_plane,
        data=np.ones((20, 10, 10)),
        unit="n.a.",
        rate=30.0,
    )

    plane_segmentation = PlaneSegmentation(
        description="my plane segmentation",
        imaging_plane=imaging_plane,
        reference_images=two_photon_series,
    )

    plane_segmentation.add_roi(image_mask=np.ones((10, 10)))

    assert check_plane_segmentation_image_mask_shape_against_ref_images(plane_segmentation) is None


def test_fail_check_plane_segmentation_image_mask_dims_against_imageseries():

    device = Device(
        name="Microscope", description="My two-photon microscope", manufacturer="The best microscope manufacturer"
    )
    optical_channel = OpticalChannel(name="OpticalChannel", description="an optical channel", emission_lambda=500.0)
    imaging_plane = ImagingPlane(
        name="ImagingPlane",
        optical_channel=optical_channel,
        imaging_rate=30.0,
        description="a very interesting part of the brain",
        device=device,
        excitation_lambda=300.0,
        indicator="GFP",
        location="V1",
        grid_spacing=[0.01, 0.01],
        grid_spacing_unit="meters",
        origin_coords=[1.0, 2.0, 3.0],
        origin_coords_unit="meters",
    )

    two_photon_series = TwoPhotonSeries(
        name="TwoPhotonSeries",
        imaging_plane=imaging_plane,
        data=np.ones((20, 10, 10)),
        unit="n.a.",
        rate=30.0,
    )

    plane_segmentation = PlaneSegmentation(
        description="my plane segmentation",
        imaging_plane=imaging_plane,
        reference_images=two_photon_series,
    )

    plane_segmentation.add_roi(image_mask=np.ones((9, 10)))

    assert check_plane_segmentation_image_mask_shape_against_ref_images(plane_segmentation) == [
        InspectorMessage(
            message="image_mask of shape (9, 10) does not match reference image TwoPhotonSeries with shape (10, 10).",
            importance=Importance.BEST_PRACTICE_VIOLATION,
            check_function_name="check_plane_segmentation_image_mask_shape_against_ref_images",
            object_type="PlaneSegmentation",
            object_name="ImagingPlane",
            location="/",
        )
    ]



"""Check functions that examine any general neurodata_type with the available attributes."""
from ..register_checks import register_check, InspectorMessage, Importance

COMMON_DESCRIPTION_PLACEHOLDERS = ["no description", "no desc", "none", "placeholder"]


@register_check(importance=Importance.CRITICAL, neurodata_type=None)
def check_name_slashes(obj):
    """Check if there  has been added for the session."""
    if hasattr(obj, "name") and any((x in obj.name for x in ["/", "\\"])):
        return InspectorMessage(message="Object name contains slashes.")


@register_check(importance=Importance.BEST_PRACTICE_SUGGESTION, neurodata_type=None)
def check_description(obj):
    """
    Check if the description is a not missing or a placeholder.

    Best Practice: :ref:`best_practice_placeholders`
    """
    if not hasattr(obj, "description"):
        return
    if obj.description is None or obj.description.strip(" ") == "":
        return InspectorMessage(message="Description is missing.")
    if obj.description.lower().strip(".") in COMMON_DESCRIPTION_PLACEHOLDERS:
        return InspectorMessage(message=f"Description ('{obj.description}') is a placeholder.")



from enum import Enum

from hdmf.common import DynamicTable
from hdmf.testing import TestCase
from pynwb import TimeSeries

from nwbinspector.register_checks import register_check, Importance, Severity, InspectorMessage


class TestRegisterClass(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.default_time_series = TimeSeries(name="temp", data=[], unit="", rate=1.0)

    def assertFunctionInChecks(self, check_function, available_checks):
        function_names = [x.__name__ for x in available_checks]
        self.assertIn(member=check_function.__name__, container=function_names)

    def test_register_importance_error_non_enum_type(self):
        bad_importance = "test_bad_importance"

        with self.assertRaisesWith(
            exc_type=ValueError,
            exc_msg=(
                f"Indicated importance ({bad_importance}) of custom check (bad_importance_function) is not a valid "
                "importance level! Please choose one of Importance.CRITICAL, Importance.BEST_PRACTICE_VIOLATION, "
                "or Importance.BEST_PRACTICE_SUGGESTION."
            ),
        ):

            @register_check(importance=bad_importance, neurodata_type=None)
            def bad_importance_function():
                pass

    def test_register_importance_error_enum_type(self):
        class SomeRandomEnum(Enum):
            random_name = -1

        bad_importance = SomeRandomEnum.random_name

        with self.assertRaisesWith(
            exc_type=ValueError,
            exc_msg=(
                f"Indicated importance ({bad_importance}) of custom check (bad_importance_function) is not a valid "
                "importance level! Please choose one of Importance.CRITICAL, Importance.BEST_PRACTICE_VIOLATION, "
                "or Importance.BEST_PRACTICE_SUGGESTION."
            ),
        ):

            @register_check(importance=bad_importance, neurodata_type=None)
            def bad_importance_function():
                pass

    def test_register_importance_error_both_forbidden_check_levels(self):
        for forbidden_importance in [Importance.ERROR, Importance.PYNWB_VALIDATION]:
            with self.assertRaisesWith(
                exc_type=ValueError,
                exc_msg=(
                    f"Indicated importance ({forbidden_importance}) of custom check (forbidden_importance_function) is "
                    "not a valid importance level! Please choose one of Importance.CRITICAL, "
                    "Importance.BEST_PRACTICE_VIOLATION, or Importance.BEST_PRACTICE_SUGGESTION."
                ),
            ):

                @register_check(importance=forbidden_importance, neurodata_type=None)
                def forbidden_importance_function():
                    pass

    def test_register_severity_error(self):
        bad_severity = "test_bad_severity"

        with self.assertRaisesWith(
            exc_type=ValueError,
            exc_msg=(
                f"Indicated severity ({bad_severity}) of custom check "
                "(bad_severity_function) is not a valid severity level! Please choose one of "
                "Severity.HIGH, Severity.LOW, or do not specify any severity."
            ),
        ):

            @register_check(importance=Importance.BEST_PRACTICE_SUGGESTION, neurodata_type=None)
            def bad_severity_function(time_series: TimeSeries):
                return InspectorMessage(severity=bad_severity, message="")

            bad_severity_function(time_series=self.default_time_series)

    def test_register_all_severity_levels(self):
        importance = Importance.BEST_PRACTICE_SUGGESTION
        for severity in Severity:

            @register_check(importance=importance, neurodata_type=TimeSeries)
            def good_check_function(time_series: TimeSeries):
                return InspectorMessage(severity=severity, message="")

            self.assertEqual(
                first=good_check_function(time_series=self.default_time_series),
                second=InspectorMessage(
                    severity=severity,
                    message="",
                    importance=importance,
                    check_function_name="good_check_function",
                    object_type="TimeSeries",
                    object_name="temp",
                    location="/",
                ),
            )

    def test_register_missing_severity(self):
        importance = Importance.BEST_PRACTICE_SUGGESTION

        @register_check(importance=importance, neurodata_type=TimeSeries)
        def good_check_function(time_series: TimeSeries):
            return InspectorMessage(message="")

        self.assertEqual(
            first=good_check_function(time_series=self.default_time_series),
            second=InspectorMessage(
                message="",
                importance=importance,
                severity=Severity.LOW,
                check_function_name="good_check_function",
                object_type="TimeSeries",
                object_name="temp",
                location="/",
            ),
        )

    def test_register_available_checks_all_importance_levels_same_neurodata_type(self):
        from nwbinspector import available_checks

        neurodata_type = DynamicTable
        for importance in [
            Importance.CRITICAL,
            Importance.BEST_PRACTICE_VIOLATION,
            Importance.BEST_PRACTICE_SUGGESTION,
        ]:

            @register_check(importance=importance, neurodata_type=neurodata_type)
            def good_check_function():
                pass

            assert good_check_function in available_checks

    def test_register_available_checks_same_importance_level_different_neurodata_types(self):
        from nwbinspector import available_checks

        importance = Importance.BEST_PRACTICE_VIOLATION
        neurodata_type_1 = DynamicTable
        neurodata_type_2 = TimeSeries

        @register_check(importance=importance, neurodata_type=neurodata_type_1)
        def good_check_function_1():
            pass

        assert good_check_function_1 in available_checks

        @register_check(importance=importance, neurodata_type=neurodata_type_2)
        def good_check_function_2():
            pass

        assert good_check_function_2 in available_checks

    def test_register_available_checks_different_importance_levels_different_neurodata_types(self):
        from nwbinspector import available_checks

        importance_1 = Importance.BEST_PRACTICE_SUGGESTION
        importance_2 = Importance.BEST_PRACTICE_VIOLATION
        neurodata_type_1 = DynamicTable
        neurodata_type_2 = TimeSeries

        @register_check(importance=importance_1, neurodata_type=neurodata_type_1)
        def good_check_function_1():
            pass

        assert good_check_function_1 in available_checks

        @register_check(importance=importance_2, neurodata_type=neurodata_type_2)
        def good_check_function_2():
            pass

        assert good_check_function_2 in available_checks



import os
import pytest
from shutil import rmtree
from tempfile import mkdtemp
from pathlib import Path
from unittest import TestCase
from datetime import datetime

import numpy as np
from pynwb import NWBFile, NWBHDF5IO, TimeSeries
from pynwb.file import TimeIntervals
from pynwb.behavior import SpatialSeries, Position
from hdmf.common import DynamicTable
from natsort import natsorted

from nwbinspector import (
    Importance,
    check_small_dataset_compression,
    check_regular_timestamps,
    check_data_orientation,
    check_timestamps_match_first_dimension,
    check_subject_exists,
    load_config,
)
from nwbinspector import inspect_all, inspect_nwb
from nwbinspector.register_checks import Severity, InspectorMessage, register_check
from nwbinspector.testing import check_streaming_tests_enabled
from nwbinspector.tools import make_minimal_nwbfile
from nwbinspector.utils import FilePathType

STREAMING_TESTS_ENABLED, DISABLED_STREAMING_TESTS_REASON = check_streaming_tests_enabled()


def add_big_dataset_no_compression(nwbfile: NWBFile):
    time_series = TimeSeries(
        name="test_time_series_1", data=np.zeros(shape=int(1.1e9 / np.dtype("float").itemsize)), rate=1.0, unit=""
    )
    nwbfile.add_acquisition(time_series)


def add_regular_timestamps(nwbfile: NWBFile):
    regular_timestamps = np.arange(1.2, 11.2, 2)
    timestamps_length = len(regular_timestamps)
    time_series = TimeSeries(
        name="test_time_series_2",
        data=np.zeros(shape=(timestamps_length, timestamps_length - 1)),
        timestamps=regular_timestamps,
        unit="",
    )
    nwbfile.add_acquisition(time_series)


def add_flipped_data_orientation_to_processing(nwbfile: NWBFile):
    spatial_series = SpatialSeries(
        name="my_spatial_series", data=np.zeros(shape=(2, 3)), reference_frame="unknown", rate=1.0
    )
    module = nwbfile.create_processing_module(name="behavior", description="")
    module.add(Position(spatial_series=spatial_series))


def add_non_matching_timestamps_dimension(nwbfile: NWBFile):
    timestamps = [1.0, 2.1, 3.0]
    timestamps_length = len(timestamps)
    time_series = TimeSeries(
        name="test_time_series_3",
        data=np.zeros(shape=(timestamps_length + 1, timestamps_length)),
        timestamps=timestamps,
        unit="",
    )
    nwbfile.add_acquisition(time_series)


def add_simple_table(nwbfile: NWBFile):
    time_intervals = TimeIntervals(name="test_table", description="desc")
    time_intervals.add_row(start_time=2.0, stop_time=3.0)
    time_intervals.add_row(start_time=1.0, stop_time=2.0)
    nwbfile.add_acquisition(time_intervals)


class TestInspector(TestCase):
    maxDiff = None

    @classmethod
    def setUpClass(cls):
        cls.tempdir = Path(mkdtemp())
        cls.checks = [
            check_small_dataset_compression,
            check_regular_timestamps,
            check_data_orientation,
            check_timestamps_match_first_dimension,
        ]
        num_nwbfiles = 3
        nwbfiles = list()
        for j in range(num_nwbfiles):
            nwbfiles.append(make_minimal_nwbfile())
        add_big_dataset_no_compression(nwbfiles[0])
        add_regular_timestamps(nwbfiles[0])
        add_flipped_data_orientation_to_processing(nwbfiles[0])
        add_non_matching_timestamps_dimension(nwbfiles[0])
        add_simple_table(nwbfiles[0])
        add_regular_timestamps(nwbfiles[1])
        # Last file to be left without violations

        cls.nwbfile_paths = [str(cls.tempdir / f"testing{j}.nwb") for j in range(num_nwbfiles)]
        for nwbfile_path, nwbfile in zip(cls.nwbfile_paths, nwbfiles):
            with NWBHDF5IO(path=nwbfile_path, mode="w") as io:
                io.write(nwbfile)

    @classmethod
    def tearDownClass(cls):
        rmtree(cls.tempdir)

    def assertFileExists(self, path: FilePathType):
        path = Path(path)
        assert path.exists()

    def assertLogFileContentsEqual(
        self, test_file_path: FilePathType, true_file_path: FilePathType, skip_first_newlines: bool = False
    ):
        with open(file=test_file_path, mode="r") as test_file:
            with open(file=true_file_path, mode="r") as true_file:
                test_file_lines = test_file.readlines()
                if skip_first_newlines:
                    for line_number, test_line in enumerate(test_file_lines):
                        if test_line != "\n":
                            skip_first_n_lines = line_number
                            break
                else:
                    skip_first_n_lines = 0
                true_file_lines = true_file.readlines()
                for line_number, test_line in enumerate(test_file_lines):
                    if "Timestamp: " in test_line:
                        # Transform the test file header to match ground true example
                        test_file_lines[line_number] = "Timestamp: 2022-04-01 13:32:13.756390-04:00\n"
                        test_file_lines[line_number + 1] = "Platform: Windows-10-10.0.19043-SP0\n"
                        test_file_lines[line_number + 2] = "NWBInspector version: 0.3.6\n"
                    if ".nwb" in test_line:
                        # Transform temporary testing path and formatted to hardcoded fake path
                        str_loc = test_line.find(".nwb")
                        correction_str = test_line.replace(test_line[5 : str_loc - 8], "./")  # noqa: E203 (black)
                        test_file_lines[line_number] = correction_str
                self.assertEqual(first=test_file_lines[skip_first_n_lines:-1], second=true_file_lines)

    def test_inspect_all(self):
        test_results = list(inspect_all(path=self.tempdir, select=[x.__name__ for x in self.checks]))
        true_results = [
            InspectorMessage(
                message="data is not compressed. Consider enabling compression when writing a dataset.",
                importance=Importance.BEST_PRACTICE_SUGGESTION,
                severity=Severity.LOW,
                check_function_name="check_small_dataset_compression",
                object_type="TimeSeries",
                object_name="test_time_series_1",
                location="/acquisition/test_time_series_1",
                file_path=self.nwbfile_paths[0],
            ),
            InspectorMessage(
                message=(
                    "TimeSeries appears to have a constant sampling rate. Consider specifying starting_time=1.2 "
                    "and rate=2.0 instead of timestamps."
                ),
                importance=Importance.BEST_PRACTICE_VIOLATION,
                severity=Severity.LOW,
                check_function_name="check_regular_timestamps",
                object_type="TimeSeries",
                object_name="test_time_series_2",
                location="/acquisition/test_time_series_2",
                file_path=self.nwbfile_paths[0],
            ),
            InspectorMessage(
                message=(
                    "Data may be in the wrong orientation. Time should be in the first dimension, and is usually "
                    "the longest dimension. Here, another dimension is longer."
                ),
                importance=Importance.CRITICAL,
                severity=Severity.LOW,
                check_function_name="check_data_orientation",
                object_type="SpatialSeries",
                object_name="my_spatial_series",
                location="/processing/behavior/Position/my_spatial_series",
                file_path=self.nwbfile_paths[0],
            ),
            InspectorMessage(
                message="The length of the first dimension of data does not match the length of timestamps.",
                importance=Importance.CRITICAL,
                severity=Severity.LOW,
                check_function_name="check_timestamps_match_first_dimension",
                object_type="TimeSeries",
                object_name="test_time_series_3",
                location="/acquisition/test_time_series_3",
                file_path=self.nwbfile_paths[0],
            ),
            InspectorMessage(
                message=(
                    "TimeSeries appears to have a constant sampling rate. Consider specifying starting_time=1.2 "
                    "and rate=2.0 instead of timestamps."
                ),
                importance=Importance.BEST_PRACTICE_VIOLATION,
                severity=Severity.LOW,
                check_function_name="check_regular_timestamps",
                object_type="TimeSeries",
                object_name="test_time_series_2",
                location="/acquisition/test_time_series_2",
                file_path=self.nwbfile_paths[1],
            ),
        ]
        self.assertCountEqual(first=test_results, second=true_results)

        def test_inspect_all_parallel(self):
            test_results = list(
                inspect_all(path=Path(self.nwbfile_paths[0]).parent, select=[x.__name__ for x in self.checks], n_jobs=2)
            )
            true_results = [
                InspectorMessage(
                    message="data is not compressed. Consider enabling compression when writing a dataset.",
                    importance=Importance.BEST_PRACTICE_SUGGESTION,
                    severity=Severity.LOW,
                    check_function_name="check_small_dataset_compression",
                    object_type="TimeSeries",
                    object_name="test_time_series_1",
                    location="/acquisition/test_time_series_1",
                    file_path=self.nwbfile_paths[0],
                ),
                InspectorMessage(
                    message=(
                        "TimeSeries appears to have a constant sampling rate. Consider specifying starting_time=1.2 "
                        "and rate=2.0 instead of timestamps."
                    ),
                    importance=Importance.BEST_PRACTICE_VIOLATION,
                    severity=Severity.LOW,
                    check_function_name="check_regular_timestamps",
                    object_type="TimeSeries",
                    object_name="test_time_series_2",
                    location="/acquisition/test_time_series_2",
                    file_path=self.nwbfile_paths[0],
                ),
                InspectorMessage(
                    message=(
                        "Data may be in the wrong orientation. Time should be in the first dimension, and is usually "
                        "the longest dimension. Here, another dimension is longer."
                    ),
                    importance=Importance.CRITICAL,
                    severity=Severity.LOW,
                    check_function_name="check_data_orientation",
                    object_type="SpatialSeries",
                    object_name="my_spatial_series",
                    location="/processing/behavior/Position/my_spatial_series",
                    file_path=self.nwbfile_paths[0],
                ),
                InspectorMessage(
                    message="The length of the first dimension of data does not match the length of timestamps.",
                    importance=Importance.CRITICAL,
                    severity=Severity.LOW,
                    check_function_name="check_timestamps_match_first_dimension",
                    object_type="TimeSeries",
                    object_name="test_time_series_3",
                    location="/acquisition/test_time_series_3",
                    file_path=self.nwbfile_paths[0],
                ),
                InspectorMessage(
                    message=(
                        "TimeSeries appears to have a constant sampling rate. Consider specifying starting_time=1.2 "
                        "and rate=2.0 instead of timestamps."
                    ),
                    importance=Importance.BEST_PRACTICE_VIOLATION,
                    severity=Severity.LOW,
                    check_function_name="check_regular_timestamps",
                    object_type="TimeSeries",
                    object_name="test_time_series_2",
                    location="/acquisition/test_time_series_2",
                    file_path=self.nwbfile_paths[1],
                ),
            ]
            self.assertCountEqual(first=test_results, second=true_results)

    def test_inspect_nwb(self):
        test_results = list(inspect_nwb(nwbfile_path=self.nwbfile_paths[0], checks=self.checks))
        true_results = [
            InspectorMessage(
                message="data is not compressed. Consider enabling compression when writing a dataset.",
                severity=Severity.LOW,
                importance=Importance.BEST_PRACTICE_SUGGESTION,
                check_function_name="check_small_dataset_compression",
                object_type="TimeSeries",
                object_name="test_time_series_1",
                location="/acquisition/test_time_series_1",
                file_path=self.nwbfile_paths[0],
            ),
            InspectorMessage(
                message=(
                    "TimeSeries appears to have a constant sampling rate. Consider specifying starting_time=1.2 and "
                    "rate=2.0 instead of timestamps."
                ),
                severity=Severity.LOW,
                importance=Importance.BEST_PRACTICE_VIOLATION,
                check_function_name="check_regular_timestamps",
                object_type="TimeSeries",
                object_name="test_time_series_2",
                location="/acquisition/test_time_series_2",
                file_path=self.nwbfile_paths[0],
            ),
            InspectorMessage(
                message=(
                    "Data may be in the wrong orientation. Time should be in the first dimension, and is usually the "
                    "longest dimension. Here, another dimension is longer."
                ),
                importance=Importance.CRITICAL,
                check_function_name="check_data_orientation",
                object_type="SpatialSeries",
                object_name="my_spatial_series",
                location="/processing/behavior/Position/my_spatial_series",
                file_path=self.nwbfile_paths[0],
            ),
            InspectorMessage(
                message="The length of the first dimension of data does not match the length of timestamps.",
                importance=Importance.CRITICAL,
                check_function_name="check_timestamps_match_first_dimension",
                object_type="TimeSeries",
                object_name="test_time_series_3",
                location="/acquisition/test_time_series_3",
                file_path=self.nwbfile_paths[0],
            ),
        ]
        self.assertCountEqual(first=test_results, second=true_results)

    def test_inspect_nwb_importance_threshold_as_importance(self):
        test_results = list(
            inspect_nwb(
                nwbfile_path=self.nwbfile_paths[0], checks=self.checks, importance_threshold=Importance.CRITICAL
            )
        )
        true_results = [
            InspectorMessage(
                message=(
                    "Data may be in the wrong orientation. Time should be in the first dimension, and is "
                    "usually the longest dimension. Here, another dimension is longer."
                ),
                importance=Importance.CRITICAL,
                check_function_name="check_data_orientation",
                object_type="SpatialSeries",
                object_name="my_spatial_series",
                location="/processing/behavior/Position/my_spatial_series",
                file_path=self.nwbfile_paths[0],
            ),
            InspectorMessage(
                message="The length of the first dimension of data does not match the length of timestamps.",
                importance=Importance.CRITICAL,
                check_function_name="check_timestamps_match_first_dimension",
                object_type="TimeSeries",
                object_name="test_time_series_3",
                location="/acquisition/test_time_series_3",
                file_path=self.nwbfile_paths[0],
            ),
        ]
        self.assertCountEqual(first=test_results, second=true_results)

    def test_inspect_nwb_importance_threshold_as_string(self):
        test_results = list(
            inspect_nwb(nwbfile_path=self.nwbfile_paths[0], checks=self.checks, importance_threshold="CRITICAL")
        )
        true_results = [
            InspectorMessage(
                message=(
                    "Data may be in the wrong orientation. Time should be in the first dimension, and is "
                    "usually the longest dimension. Here, another dimension is longer."
                ),
                importance=Importance.CRITICAL,
                check_function_name="check_data_orientation",
                object_type="SpatialSeries",
                object_name="my_spatial_series",
                location="/processing/behavior/Position/my_spatial_series",
                file_path=self.nwbfile_paths[0],
            ),
            InspectorMessage(
                message="The length of the first dimension of data does not match the length of timestamps.",
                importance=Importance.CRITICAL,
                check_function_name="check_timestamps_match_first_dimension",
                object_type="TimeSeries",
                object_name="test_time_series_3",
                location="/acquisition/test_time_series_3",
                file_path=self.nwbfile_paths[0],
            ),
        ]
        self.assertCountEqual(first=test_results, second=true_results)

    def test_command_line_runs_cli_only(self):
        console_output_file = self.tempdir / "test_console_output.txt"
        os.system(
            f"nwbinspector {str(self.tempdir)} --overwrite --select check_timestamps_match_first_dimension,"
            "check_data_orientation,check_regular_timestamps,check_small_dataset_compression"
            f"> {console_output_file}"
        )
        self.assertLogFileContentsEqual(
            test_file_path=console_output_file,
            true_file_path=Path(__file__).parent / "true_nwbinspector_default_report.txt",
            skip_first_newlines=True,
        )

    def test_command_line_runs_cli_only_parallel(self):
        console_output_file = self.tempdir / "test_console_output_2.txt"
        os.system(
            f"nwbinspector {str(self.tempdir)} --n-jobs 2 --overwrite --select check_timestamps_match_first_dimension,"
            "check_data_orientation,check_regular_timestamps,check_small_dataset_compression"
            f"> {console_output_file}"
        )
        self.assertLogFileContentsEqual(
            test_file_path=console_output_file,
            true_file_path=Path(__file__).parent / "true_nwbinspector_default_report.txt",
            skip_first_newlines=True,
        )

    def test_command_line_saves_report(self):
        console_output_file = self.tempdir / "test_console_output_3.txt"
        os.system(
            f"nwbinspector {str(self.nwbfile_paths[0])} "
            f"--report-file-path {self.tempdir / 'test_nwbinspector_report_1.txt'}"
            f"> {console_output_file}"
        )
        self.assertFileExists(path=self.tempdir / "test_nwbinspector_report_1.txt")

    def test_command_line_organizes_levels(self):
        console_output_file = self.tempdir / "test_console_output_4.txt"
        os.system(
            f"nwbinspector {str(self.nwbfile_paths[0])} "
            f"--report-file-path {self.tempdir / 'test_nwbinspector_report_2.txt'} "
            "--levels importance,check_function_name,file_path"
            f"> {console_output_file}"
        )
        self.assertFileExists(path=self.tempdir / "test_nwbinspector_report_2.txt")

    def test_command_line_runs_saves_json(self):
        json_fpath = self.tempdir / "nwbinspector_results.json"
        os.system(f"nwbinspector {str(self.nwbfile_paths[0])} --json-file-path {json_fpath}")
        self.assertFileExists(path=json_fpath)

    def test_command_line_on_directory_matches_file(self):
        console_output_file = self.tempdir / "test_console_output_5.txt"
        os.system(
            f"nwbinspector {str(self.tempdir)} --overwrite --select check_timestamps_match_first_dimension,"
            "check_data_orientation,check_regular_timestamps,check_small_dataset_compression"
            f" --report-file-path {self.tempdir / 'test_nwbinspector_report_3.txt'}"
            f"> {console_output_file}"
        )
        self.assertLogFileContentsEqual(
            test_file_path=self.tempdir / "test_nwbinspector_report_3.txt",
            true_file_path=Path(__file__).parent / "true_nwbinspector_default_report.txt",
            skip_first_newlines=True,
        )

    def test_iterable_check_function(self):
        @register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=DynamicTable)
        def iterable_check_function(table: DynamicTable):
            for col in table.columns:
                yield InspectorMessage(message=f"Column: {col.name}")

        test_results = list(inspect_nwb(nwbfile_path=self.nwbfile_paths[0], select=["iterable_check_function"]))
        true_results = [
            InspectorMessage(
                message="Column: start_time",
                importance=Importance.BEST_PRACTICE_VIOLATION,
                check_function_name="iterable_check_function",
                object_type="TimeIntervals",
                object_name="test_table",
                file_path=self.nwbfile_paths[0],
            ),
            InspectorMessage(
                message="Column: stop_time",
                importance=Importance.BEST_PRACTICE_VIOLATION,
                check_function_name="iterable_check_function",
                object_type="TimeIntervals",
                object_name="test_table",
                file_path=self.nwbfile_paths[0],
            ),
        ]
        self.assertCountEqual(first=test_results, second=true_results)

    def test_inspect_nwb_manual_iteration(self):
        generator = inspect_nwb(nwbfile_path=self.nwbfile_paths[0], checks=self.checks)
        message = next(generator)
        true_result = InspectorMessage(
            message="data is not compressed. Consider enabling compression when writing a dataset.",
            importance=Importance.BEST_PRACTICE_SUGGESTION,
            severity=Severity.LOW,
            check_function_name="check_small_dataset_compression",
            object_type="TimeSeries",
            object_name="test_time_series_1",
            location="/acquisition/test_time_series_1",
            file_path=self.nwbfile_paths[0],
        )
        self.assertEqual(message, true_result)

    def test_inspect_nwb_manual_iteration_stop(self):
        generator = inspect_nwb(nwbfile_path=self.nwbfile_paths[2], checks=self.checks)
        with self.assertRaises(expected_exception=StopIteration):
            next(generator)

    def test_inspect_nwb_dandi_config(self):
        config_checks = [check_subject_exists] + self.checks
        test_results = list(
            inspect_nwb(
                nwbfile_path=self.nwbfile_paths[0],
                checks=config_checks,
                config=load_config(filepath_or_keyword="dandi"),
            )
        )
        true_results = [
            InspectorMessage(
                message="Subject is missing.",
                importance=Importance.CRITICAL,  # Normally a BEST_PRACTICE_SUGGESTION
                check_function_name="check_subject_exists",
                object_type="NWBFile",
                object_name="root",
                location="/",
                file_path=self.nwbfile_paths[0],
            ),
            InspectorMessage(
                message=(
                    "Data may be in the wrong orientation. "
                    "Time should be in the first dimension, and is usually the longest dimension. "
                    "Here, another dimension is longer."
                ),
                importance=Importance.BEST_PRACTICE_VIOLATION,  # Normally CRITICAL, now a BEST_PRACTICE_VIOLATION
                check_function_name="check_data_orientation",
                object_type="SpatialSeries",
                object_name="my_spatial_series",
                location="/processing/behavior/Position/my_spatial_series",
                file_path=self.nwbfile_paths[0],
            ),
            InspectorMessage(
                message="data is not compressed. Consider enabling compression when writing a dataset.",
                importance=Importance.BEST_PRACTICE_SUGGESTION,
                check_function_name="check_small_dataset_compression",
                object_type="TimeSeries",
                object_name="test_time_series_1",
                location="/acquisition/test_time_series_1",
                file_path=self.nwbfile_paths[0],
            ),
            InspectorMessage(
                message=(
                    "TimeSeries appears to have a constant sampling rate. "
                    "Consider specifying starting_time=1.2 and rate=2.0 instead of timestamps."
                ),
                importance=Importance.BEST_PRACTICE_VIOLATION,
                check_function_name="check_regular_timestamps",
                object_type="TimeSeries",
                object_name="test_time_series_2",
                location="/acquisition/test_time_series_2",
                file_path=self.nwbfile_paths[0],
            ),
            InspectorMessage(
                message="The length of the first dimension of data does not match the length of timestamps.",
                importance=Importance.CRITICAL,
                check_function_name="check_timestamps_match_first_dimension",
                object_type="TimeSeries",
                object_name="test_time_series_3",
                location="/acquisition/test_time_series_3",
                file_path=self.nwbfile_paths[0],
            ),
        ]
        self.assertCountEqual(first=test_results, second=true_results)


@pytest.mark.skipif(not STREAMING_TESTS_ENABLED, reason=DISABLED_STREAMING_TESTS_REASON or "")
def test_dandiset_streaming():
    messages = list(inspect_all(path="000126", select=["check_subject_species_exists"], stream=True))
    assert messages[0] == InspectorMessage(
        message="Subject species is missing.",
        importance=Importance.BEST_PRACTICE_VIOLATION,
        check_function_name="check_subject_species_exists",
        object_type="Subject",
        object_name="subject",
        location="/general/subject",
        file_path="sub-1/sub-1.nwb",
    )


@pytest.mark.skipif(not STREAMING_TESTS_ENABLED, reason=DISABLED_STREAMING_TESTS_REASON or "")
def test_dandiset_streaming_parallel():
    messages = list(inspect_all(path="000126", select=["check_subject_species_exists"], stream=True, n_jobs=2))
    assert messages[0] == InspectorMessage(
        message="Subject species is missing.",
        importance=Importance.BEST_PRACTICE_VIOLATION,
        check_function_name="check_subject_species_exists",
        object_type="Subject",
        object_name="subject",
        location="/general/subject",
        file_path="sub-1/sub-1.nwb",
    )


@pytest.mark.skipif(not STREAMING_TESTS_ENABLED, reason=DISABLED_STREAMING_TESTS_REASON or "")
class TestStreamingCLI(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.tempdir = Path(mkdtemp())

    @classmethod
    def tearDownClass(cls):
        rmtree(cls.tempdir)

    def assertFileExists(self, path: FilePathType):
        path = Path(path)
        assert path.exists()

    def test_dandiset_streaming_cli(self):
        console_output_file = self.tempdir / "test_console_streaming_output_1.txt"
        os.system(
            f"nwbinspector 000126 --stream "
            f"--report-file-path {self.tempdir / 'test_nwbinspector_streaming_report_6.txt'}"
            f"> {console_output_file}"
        )
        self.assertFileExists(path=self.tempdir / "test_nwbinspector_streaming_report_6.txt")

    def test_dandiset_streaming_cli_parallel(self):
        console_output_file = self.tempdir / "test_console_streaming_output_2.txt"
        os.system(
            f"nwbinspector https://dandiarchive.org/dandiset/000126/0.210813.0327 --stream --n-jobs 2 "
            f"--report-file-path {self.tempdir / 'test_nwbinspector_streaming_report_7.txt'}"
            f"> {console_output_file}"
        )
        self.assertFileExists(path=self.tempdir / "test_nwbinspector_streaming_report_7.txt")


class TestCheckUniqueIdentifiersPass(TestCase):
    maxDiff = None

    @classmethod
    def setUpClass(cls):
        cls.tempdir = Path(mkdtemp())
        num_nwbfiles = 3
        unique_id_nwbfiles = list()
        for j in range(num_nwbfiles):
            unique_id_nwbfiles.append(make_minimal_nwbfile())

        cls.unique_id_nwbfile_paths = [str(cls.tempdir / f"unique_id_testing{j}.nwb") for j in range(num_nwbfiles)]
        for nwbfile_path, nwbfile in zip(cls.unique_id_nwbfile_paths, unique_id_nwbfiles):
            with NWBHDF5IO(path=nwbfile_path, mode="w") as io:
                io.write(nwbfile)

    @classmethod
    def tearDownClass(cls):
        rmtree(cls.tempdir)

    def test_check_unique_identifiers_pass(self):
        assert list(inspect_all(path=self.tempdir, select=["check_data_orientation"])) == []


class TestCheckUniqueIdentifiersFail(TestCase):
    maxDiff = None

    @classmethod
    def setUpClass(cls):
        cls.tempdir = Path(mkdtemp())
        num_nwbfiles = 3
        non_unique_id_nwbfiles = list()
        for j in range(num_nwbfiles):
            non_unique_id_nwbfiles.append(
                NWBFile(
                    session_description="",
                    identifier="not a unique identifier!",
                    session_start_time=datetime.now().astimezone(),
                )
            )

        cls.non_unique_id_nwbfile_paths = [
            str(cls.tempdir / f"non_unique_id_testing{j}.nwb") for j in range(num_nwbfiles)
        ]
        for nwbfile_path, nwbfile in zip(cls.non_unique_id_nwbfile_paths, non_unique_id_nwbfiles):
            with NWBHDF5IO(path=nwbfile_path, mode="w") as io:
                io.write(nwbfile)

    @classmethod
    def tearDownClass(cls):
        rmtree(cls.tempdir)

    def test_check_unique_identifiers_fail(self):
        assert list(inspect_all(path=self.tempdir, select=["check_data_orientation"])) == [
            InspectorMessage(
                message=(
                    "The identifier 'not a unique identifier!' is used across the .nwb files: "
                    f"{natsorted([Path(x).name for x in self.non_unique_id_nwbfile_paths])}. "
                    "The identifier of any NWBFile should be a completely unique value - "
                    "we recommend using uuid4 to achieve this."
                ),
                importance=Importance.CRITICAL,
                check_function_name="check_unique_identifiers",
                object_type="NWBFile",
                object_name="root",
                location="/",
                file_path=str(self.tempdir),
            )
        ]



from pathlib import Path
from tempfile import mkdtemp
from shutil import rmtree
from unittest import TestCase
from datetime import datetime

import h5py
import numpy as np
from pynwb import NWBContainer, NWBFile

from nwbinspector import (
    InspectorMessage,
    Importance,
    check_small_dataset_compression,
    check_large_dataset_compression,
    check_empty_string_for_optional_attribute,
)
from nwbinspector.register_checks import Severity


class TestNWBContainers(TestCase):
    def setUp(self):
        self.test_folder = Path(mkdtemp())
        self.file_path = str(self.test_folder / "test_file.nwb")

    def tearDown(self):
        rmtree(self.test_folder)

    @staticmethod
    def add_dataset_to_nwb_container(file: h5py.File, gb_size: float):
        dataset = file.create_dataset(
            name="test_dataset",
            data=np.zeros(shape=int(gb_size * 1e9 / np.dtype("float").itemsize)),
        )
        nwb_container = NWBContainer(name="test_container")
        nwb_container.fields.update(dataset=dataset)
        return nwb_container

    def test_check_small_dataset_compression_below_default_threshold(self):
        with h5py.File(name=self.file_path, mode="w") as file:
            nwb_container = self.add_dataset_to_nwb_container(file=file, gb_size=0.1)
            true_output = InspectorMessage(
                message="test_dataset is not compressed. Consider enabling compression when writing a dataset.",
                importance=Importance.BEST_PRACTICE_SUGGESTION,
                check_function_name="check_small_dataset_compression",
                object_type="NWBContainer",
                object_name="test_container",
                location="/",
            )
            self.assertEqual(first=check_small_dataset_compression(nwb_container=nwb_container), second=true_output)

    def test_check_small_dataset_compression_below_50MB(self):
        with h5py.File(name=self.file_path, mode="w") as file:
            nwb_container = self.add_dataset_to_nwb_container(file=file, gb_size=0.001)
            self.assertIsNone(obj=check_large_dataset_compression(nwb_container=nwb_container))

    def test_check_small_dataset_compression_below_manual_threshold(self):
        with h5py.File(name=self.file_path, mode="w") as file:
            nwb_container = self.add_dataset_to_nwb_container(file=file, gb_size=0.1)
            true_output = InspectorMessage(
                message="test_dataset is not compressed. Consider enabling compression when writing a dataset.",
                importance=Importance.BEST_PRACTICE_SUGGESTION,
                check_function_name="check_small_dataset_compression",
                object_type="NWBContainer",
                object_name="test_container",
                location="/",
            )
            self.assertEqual(
                first=check_small_dataset_compression(nwb_container=nwb_container, gb_severity_threshold=0.15),
                second=true_output,
            )

    def test_check_small_dataset_compression_above_manual_threshold(self):
        with h5py.File(name=self.file_path, mode="w") as file:
            nwb_container = self.add_dataset_to_nwb_container(file=file, gb_size=0.2)
            true_output = InspectorMessage(
                severity=Severity.HIGH,
                message="test_dataset is not compressed. Consider enabling compression when writing a dataset.",
                importance=Importance.BEST_PRACTICE_SUGGESTION,
                check_function_name="check_small_dataset_compression",
                object_type="NWBContainer",
                object_name="test_container",
                location="/",
            )
            self.assertEqual(
                first=check_small_dataset_compression(nwb_container=nwb_container, gb_severity_threshold=0.15),
                second=true_output,
            )

    def test_check_large_dataset_compression_below_20GB(self):
        with h5py.File(name=self.file_path, mode="w") as file:
            nwb_container = self.add_dataset_to_nwb_container(file=file, gb_size=0.001)
            self.assertIsNone(obj=check_large_dataset_compression(nwb_container=nwb_container))


def test_hit_check_empty_string_for_optional_attribute():
    nwb = NWBFile("aa", "aa", datetime.now(), pharmacology="")

    assert check_empty_string_for_optional_attribute(nwb)[0] == InspectorMessage(
        message='The attribute "pharmacology" is optional and you have supplied an empty string. Improve my omitting '
        "this attribute (in MatNWB or PyNWB) or entering as None (in PyNWB)",
        importance=Importance.BEST_PRACTICE_SUGGESTION,
        location="/",
        object_type="NWBFile",
        object_name="root",
        check_function_name="check_empty_string_for_optional_attribute",
    )


def test_miss_check_empty_string_for_optional_attribute():
    nwb = NWBFile("aa", "aa", datetime.now())
    assert check_empty_string_for_optional_attribute(nwb) is None



from nwbinspector.testing import load_testing_config, update_testing_config


def test_update_testing_config():
    test_key = "LOCAL_PATH"
    test_value = "/a/new/path/"

    try:
        initial_testing_config = load_testing_config()
        NO_CONFIG = False  # Depending on the method of installation, a config may not have generated
    except FileNotFoundError:
        NO_CONFIG = True

    if not NO_CONFIG:
        try:
            update_testing_config(key=test_key, value=test_value)
            updated_testing_config = load_testing_config()
        finally:
            # Restore
            update_testing_config(key=test_key, value=initial_testing_config[test_key])

        assert updated_testing_config[test_key] != initial_testing_config[test_key]
        assert updated_testing_config[test_key] == test_value



"""Helper functions for internal use that rely on external dependencies (i.e., pynwb)."""
import re
from uuid import uuid4
from datetime import datetime
from typing import Optional, Dict, Tuple
from concurrent.futures import ProcessPoolExecutor, as_completed
from urllib import request
from warnings import warn

import h5py
from pynwb import NWBFile

from .utils import is_module_installed, calculate_number_of_cpu


def make_minimal_nwbfile():
    """Most basic NWBFile that can exist."""
    return NWBFile(session_description="", identifier=str(uuid4()), session_start_time=datetime.now().astimezone())


def all_of_type(nwbfile: NWBFile, neurodata_type):
    """Iterate over all objects inside an NWBFile object and return those that match the given neurodata_type."""
    for obj in nwbfile.objects.values():
        if isinstance(obj, neurodata_type):
            yield obj


def get_nwbfile_path_from_internal_object(obj):
    """Determine the file path on disk for a NWBFile given only an internal object of that file."""
    if isinstance(obj, NWBFile):
        return obj.container_source
    return obj.get_ancestor("NWBFile").container_source


def get_s3_urls_and_dandi_paths(dandiset_id: str, version_id: Optional[str] = None, n_jobs: int = 1) -> Dict[str, str]:
    """
    Collect S3 URLS from a DANDISet ID.

    Returns dictionary that maps each S3 url to the displayed file path on the DANDI archive content page.
    """
    assert is_module_installed(module_name="dandi"), "You must install DANDI to get S3 paths (pip install dandi)."
    from dandi.dandiapi import DandiAPIClient

    assert re.fullmatch(
        pattern="^[0-9]{6}$", string=dandiset_id
    ), "The specified 'path' is not a proper DANDISet ID. It should be a six-digit numeric identifier."

    s3_urls_to_dandi_paths = dict()
    n_jobs = calculate_number_of_cpu(requested_cpu=n_jobs)
    if n_jobs != 1:
        with DandiAPIClient() as client:
            dandiset = client.get_dandiset(dandiset_id=dandiset_id, version_id=version_id)
            max_workers = n_jobs if n_jobs > 0 else None
            with ProcessPoolExecutor(max_workers=max_workers) as executor:
                futures = []
                for asset in dandiset.get_assets():
                    if asset.path.split(".")[-1] == "nwb":
                        futures.append(
                            executor.submit(
                                _get_content_url_and_path, asset=asset, follow_redirects=1, strip_query=True
                            )
                        )
                    for future in as_completed(futures):
                        s3_urls_to_dandi_paths.update(future.result())
    else:
        with DandiAPIClient() as client:
            dandiset = client.get_dandiset(dandiset_id=dandiset_id, version_id=version_id)
            for asset in dandiset.get_assets():
                if asset.path.split(".")[-1] == "nwb":
                    s3_urls_to_dandi_paths.update(_get_content_url_and_path(asset=asset))
    return s3_urls_to_dandi_paths


def _get_content_url_and_path(asset, follow_redirects: int = 1, strip_query: bool = True) -> Dict[str, str]:
    """
    Private helper function for parallelization in 'get_s3_urls_and_dandi_paths'.

    Must be globally defined (not as a part of get_s3_urls..) in order to be pickled.
    """
    return {asset.get_content_url(follow_redirects=1, strip_query=True): asset.path}


def check_streaming_enabled() -> Tuple[bool, Optional[str]]:
    """
    General purpose helper for determining if the environment can support S3 DANDI streaming.

    Returns the boolean status of the check and, if False, provides a string reason for the failure for the user to
    utilize as they please (raise an error or warning with that message, print it, or ignore it).
    """
    try:
        request.urlopen("https://dandiarchive.s3.amazonaws.com/ros3test.nwb", timeout=1)
    except request.URLError:
        return False, "Internet access to DANDI failed."
    if "ros3" not in h5py.registered_drivers():
        return False, "ROS3 driver not installed."
    return True, None



"""Checks for types belonging to the pynwb.behavior module."""
import numpy as np
from pynwb.behavior import SpatialSeries, CompassDirection

from ..register_checks import register_check, Importance, InspectorMessage


@register_check(importance=Importance.CRITICAL, neurodata_type=SpatialSeries)
def check_spatial_series_dims(spatial_series: SpatialSeries):
    """Check if a SpatialSeries has the correct dimensions."""
    if len(spatial_series.data.shape) > 1 and spatial_series.data.shape[1] > 3:
        return InspectorMessage(
            message="SpatialSeries should have 1 column (x), 2 columns (x, y), or 3 columns (x, y, z)."
        )


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=CompassDirection)
def check_compass_direction_unit(compass_direction: CompassDirection):
    for spatial_series in compass_direction.spatial_series.values():
        if spatial_series.unit not in ("degrees", "radians"):
            yield InspectorMessage(
                message=f"SpatialSeries objects inside a CompassDirection object should be angular and should have a "
                f"unit of 'degrees' or 'radians', but '{spatial_series.name}' has units '{spatial_series.unit}'."
            )


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=SpatialSeries)
def check_spatial_series_radians_magnitude(spatial_series: SpatialSeries, nelems: int = 200):
    if spatial_series.unit in ("radian", "radians"):
        data = spatial_series.data[:nelems]
        if np.any(data > (2 * np.pi)) or np.any(data < (-2 * np.pi)):
            return InspectorMessage(
                message="SpatialSeries with units of radians must have values between -2pi and 2pi."
            )


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=SpatialSeries)
def check_spatial_series_degrees_magnitude(spatial_series: SpatialSeries, nelems: int = 200):
    if spatial_series.unit in ("degree", "degrees"):
        data = spatial_series.data[:nelems]
        if np.any(data > 360) or np.any(data < -360):
            return InspectorMessage(
                message="SpatialSeries with units of degrees must have values between -360 and 360."
            )



from pynwb.behavior import SpatialSeries, CompassDirection
import numpy as np

from nwbinspector import InspectorMessage, Importance
from nwbinspector.checks.behavior import (
    check_compass_direction_unit,
    check_spatial_series_dims,
    check_spatial_series_degrees_magnitude,
    check_spatial_series_radians_magnitude,
)


def test_check_spatial_series_dims():

    spatial_series = SpatialSeries(
        name="SpatialSeries",
        description="description",
        data=np.ones((10, 4)),
        rate=3.0,
        reference_frame="reference_frame",
    )
    assert check_spatial_series_dims(spatial_series) == InspectorMessage(
        message="SpatialSeries should have 1 column (x), 2 columns (x, y), or 3 columns (x, y, z).",
        importance=Importance.CRITICAL,
        check_function_name="check_spatial_series_dims",
        object_type="SpatialSeries",
        object_name="SpatialSeries",
        location="/",
    )


def test_pass_check_spatial_series_dims():

    spatial_series = SpatialSeries(
        name="SpatialSeries",
        description="description",
        data=np.ones((10, 3)),
        rate=3.0,
        reference_frame="reference_frame",
    )

    assert check_spatial_series_dims(spatial_series) is None


def test_pass_check_spatial_series_dims_1d():

    spatial_series = SpatialSeries(
        name="SpatialSeries",
        description="description",
        data=np.ones((10,)),
        rate=3.0,
        reference_frame="reference_frame",
    )

    assert check_spatial_series_dims(spatial_series) is None


def test_trigger_check_compass_direction_unit():
    obj = CompassDirection(
        spatial_series=SpatialSeries(
            name="SpatialSeries",
            description="description",
            data=np.ones((10,)),
            rate=3.0,
            reference_frame="reference_frame",
        )
    )

    assert (
        check_compass_direction_unit(obj)[0].message == f"SpatialSeries objects inside a CompassDirection object "
        f"should be angular and should have a unit of 'degrees' or 'radians', but 'SpatialSeries' has units 'meters'."
    )


def test_pass_check_compass_direction_unit():
    for unit in ("radians", "degrees"):
        obj = CompassDirection(
            spatial_series=SpatialSeries(
                name="SpatialSeries",
                description="description",
                data=np.ones((10,)),
                rate=3.0,
                reference_frame="reference_frame",
                unit=unit,
            )
        )

        assert check_compass_direction_unit(obj) is None


def test_pass_check_spatial_series_degrees_magnitude():

    spatial_series = SpatialSeries(
        name="SpatialSeries",
        description="description",
        data=np.ones((10,)),
        rate=3.0,
        reference_frame="reference_frame",
        unit="degrees",
    )

    assert check_spatial_series_degrees_magnitude(spatial_series) is None


def test_check_spatial_series_degrees_magnitude():

    spatial_series = SpatialSeries(
        name="SpatialSeries",
        description="description",
        data=np.ones((10,)) * 400,
        rate=3.0,
        reference_frame="reference_frame",
        unit="degrees",
    )

    assert check_spatial_series_degrees_magnitude(spatial_series) == InspectorMessage(
        check_function_name="check_spatial_series_degrees_magnitude",
        message="SpatialSeries with units of degrees must have values between -360 and 360.",
        importance=Importance.BEST_PRACTICE_VIOLATION,
        object_name="SpatialSeries",
        location="/",
        object_type="SpatialSeries",
    )


def test_pass_check_spatial_series_radians_magnitude():

    spatial_series = SpatialSeries(
        name="SpatialSeries",
        description="description",
        data=np.ones((10,)),
        rate=3.0,
        reference_frame="reference_frame",
        unit="radians",
    )

    assert check_spatial_series_radians_magnitude(spatial_series) is None


def test_check_spatial_series_radians_magnitude():

    spatial_series = SpatialSeries(
        name="SpatialSeries",
        description="description",
        data=np.ones((10,)) * 400,
        rate=3.0,
        reference_frame="reference_frame",
        unit="radians",
    )

    assert check_spatial_series_radians_magnitude(spatial_series) == InspectorMessage(
        check_function_name="check_spatial_series_radians_magnitude",
        message="SpatialSeries with units of radians must have values between -2pi and 2pi.",
        importance=Importance.BEST_PRACTICE_VIOLATION,
        object_name="SpatialSeries",
        location="/",
        object_type="SpatialSeries",
    )
